/*****************************************************************************/
/*
                                  mondesi.c

                        Monitor de Sistema (espaol)

MONDESI (pronounced "mon-deh-sih") is a browser-based VMS system monitor with a
sufficiently small resource footprint as not to impact system performance
(minimal CPU consumption and only one or two buffered IO per client update). 
No, the name's etiology has little significance at all (a small homage at
most).  I was just looking for a short (and cute) name.

This utility is intended as a general tool for system observation.  It is not a
troubleshooting tool as such, especially when a system is beginning to suffer
from resource exhaustion.  Not only does the data collection script need to
continue processing effectively (see 'SUFFICIENT PRIORITY' below) there is also
a Web server in the data transfer path.

MONDESI should work with all relatively modern browsers.  Version 3.n was
developed against Chrome 23, Firefox 17, MSIE 9 & 10, Opera 10, and Safari 6,
on OS X and Windows 7 platforms.  The intense graphing is CPU and memory
hungry so I'd not recommend Mozilla on an DEC 3000 (or even a PWS 500 :-)  Of
course YMMV.

MONDESI should also work with HTML5 WebSocket capable browsers.

Uses the system service $GETRMI() available for VMS V7.3 and later.  $GETRMI
seems to have been intended as the 'modern' and documented version of the
$GETSPI functionality.  Unfortunately, after dealing with it for a while, it's
difficult to shake the impression that someone lost interest halfway through
the project!  Note that V7.3-1 (and earlier) do not support some data items. 
These are shown as question marks.

Displayed values have been validated using the interactive MONITOR utility.

As much of the functionality is pushed off to the browser MONDESI relies
heavily on JavaScript contained in MONDESI.JS and WZ_JSGRAPHICS.JS (see below).

As MONDESI provides a readily accessable insight into a system (albeit,
fairly basic resource consumption information only) perhaps it might be best
available only under authorization.

Refresh rate defaults to two seconds.  To change this (less than two is not
recommended) define a logical name containing the value in seconds.

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_REFRESH 5

If MonDeSi stops updating the browser for any reason (e.g. server shutdown,
network connection interrupted) for a period of ninety seconds or more, the
node notes show a bold "STALLED".  MonDeSi periodically attempts to reestablish
lost network connections automatically.  The "STALLED" will disappear once
updates recommence.

NOTE: Orignally MONDESI was a vehicle for experimentation rather than an
application designed and built to address a specific need.  Over the years it
has received some use in anger and undergone a couple of major cycles of
enhancement.  It is now - for what it tries to provide - a capable and useful
tool to snapshot resource utilisation on a VMS platform.


MONITOR DISPLAY
---------------
A node is displayed as a text panel on the left and nine bar graphs on the
right.  These display the instantaneous (current) values for the various
resources and are updated dynamically at the interval when the data collector
script detects a change in values.  Each bar graph section contains a number
representing the current value and a number in the far right the maximum value
reached while monitoring.  All values are average per-second where applicable
(e.g. BIO, DIO, lock, etc.)

  +--------------------------------------------------------------------------+
  | NODENAME                     [_CPU_____________________:___|_______100_] |
  | 04-JAN-2013 05:20            [_MEM_____________|__________________1.5G_] |
  | Digital Personal Workstn     [_FLTS_________________________________10_] |
  | 1 CPU 1.5 GB Alpha V8.3      [_BIO_____:_|__________________________16_] |
  | Up: 44 14:29 Processes:35    [_DIO______________|____________________5_] |
  | interactive:4 batch:0        [_LOCK___|_____________________________39_] |
  | network:12 (v3.0.0/WASD)     [_NI________________________|_______15.6k_] |
  |                              [_DISK_:_________|_____________________12_] |
  | [hi][cl][di][pr][re][ze][pa] [_PRC__|_______________________________35_] |
  +--------------------------------------------------------------------------+

CPU: shows total CPU utilisation in green, non-user-mode CPU
(kernel+interrupt+etc.) in brown (usually a subset of total) and average CPU
utilisation in yellow.  Where a system has single CPU the graph range is 100%,
with two CPUs the range is 200%, three is 300%, etc.

MEM: displays system physical memory (RAM) in Mbytes.  The value in the far
right is the total.  The current used is show in green.  Used memory is
calculate by subtracting the current free list plus modified list sizes from
the total physical memory.

FLTS: provides total page faults per second in green with hard page faults as a
subset in brown.  Average page faulting for the monitor period is shown in
yellow.

BIO: displays buffered I/Os per second.  Average is shown in yellow.

DIO: displays direct I/Os per second.  Average is shown in yellow.

LOCK: provides locking activity as a total of ENQs plus DEQs plus waits. 
Locks waits are shown in brown.  Average activity is shown in yellow.

NI: network interface(s) bytes received and transmitted per second; total in
green, received in brown, transmitted data being the difference.  This includes
all data moving through the physical network interface(s) including SCS, MSCP,
etc., not just protocols such as TCP/IP. 

DISK: displays total disk I/Os in green with a subset of disk writes in
brown (disk reads being the difference).  Average activity shown in yellow.

Control Buttons:

[history] button reveals/hides the history line graphs (see below).

[cluster] enabled if a clustered node, button reveals/hides cluster data

[disk] enabled if disk monitoring configured, reveals/hides disk data

[proc] enable if process monitoring is enabled, reveals/hides process display

[scale] adjusts the y axis to the current maximum data value

[zero] button can be used to reset the collected maximum and average data.

[print] suspends updates and invokes the browser print dialog
        button shouldn't need to be hit again after print dialogue concluded

[pause][release] stops display updates while continuing to collect data
                 release recommences display updates with paused data
                 pause may be used to briefly examine something of interest
                 if paused five mins MonDeSi stops itself and shows [START]

When cluster button reveals cluster data the following additional bar graphs
and associated history graphs become visible.

  +--------------------------------------------------------------------------+
  :                                                                          :
  |                              [_SCSMSG_________________________|______1_] |
  |                              [_SCSKB__:_________|____________________5_] |
  | [hi][cl][di][pr][re][ze][pa] [_MSCP____________|_____________________4_] |
  +--------------------------------------------------------------------------+

SCSMG: count of SCS messages exchanged.  Displays total in green and received
in brown.  (MONITOR> SCS /ITEM=(M_SEND,M_RECEIVE) analogue.)  Average is
displayed in yellow. 

SCSKB: kilobyte map rate.  Displays total in green.  (MONITOR> SCS
/ITEM=(KB_MAP) analogue.)  Average is displayed in yellow. 

MSCP: displays total requests from remote systems in green and a subset showing
write requests in brown.  Average is displayed in yellow.

When the disk button reveals disk data (configurable - see below) it displays
the per-disk operations per second.

  +--------------------------------------------------------------------------+
  :                                                                          :
  |                              [_$1$DKA0:_ALPHASYS_____:________|_____30_] |
  |                              [_$1$DKA100:_BACKUP_____________________0_] |
  | [hi][cl][di][pr][re][ze][pa] [_$1$MDA0:_RAMDISK_:__|________________12_] |
  +--------------------------------------------------------------------------+

For configured disks the bar graph (and history) shows the device name and
volume name if mounted.  If not mounted only the device name.


NETWORK STATISTICS
------------------
MONDESI will scan the system for network devices (commonly ethernet) and use
any found for NI statistics.  If the system admin wishes to tailor the source
of the statistics the multi-valued logical name MONDESI_NI can be used to
specify one or more network interfaces.  Define this logical name system-wide
(or in a script wrapper procedure) as required.

For example:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_NI EWA0:,EWB0:

This logical name must exist as MONDESI initialises.  If undefined or the
specified device(s) cannot be accessed the ethernet data are greyed-out.


DISK STATISTICS
---------------
MONDESI will scan selected disk devices displaying the operations per second
for mounted devices.  To enable this element the system admin needs to define a
multi-valued logical name containing the devices to be monitored.  Those
selected for display must be defined using the device names as shown by a $
SHOW DEVICE command.

For example:

  $ SHOW DEVICE DKA

  Device                Device        Error    Volume         Free  Trans Mnt
   Name                 Status        Count     Label         Space Count Cnt
  $1$DKA0:    (KLAATU)  Mounted           0  ALPHASYS       90.22GB   597   1
  $1$DKA100:  (KLAATU)  Mounted           0  BACKUP        208.86GB    10   1

For example:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_DISK $1$DKA0:,$1$DKA100:,$1$MDA0:

This logical name must exist as MONDESI initialises.


PROCESS DISPLAY
---------------
MONDESI can be configured to display the top five processes for each of the
categories; CPU time, super+exec+kernel time, page faults, buffered and direct
I/O.  It is displayed as a table immediately below the bar graph area and above
the history area.  Generating and then transmitting the data is RELATIVELY
EXPENSIVE compared to other MonDeSi data.  In the WebSocket environment the
process data is not generated unless a client is displaying it. The collection
code automatically switches between newer and older SPI data and data structure
(coincident with VMS V8.3?).  The S+E+K data item is unavailable with the
older.

For example:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_PROC *

This logical name must exist as MONDESI initialises.


HISTORY DISPLAY
---------------
Recent data are stored within the browser and used to generate line graphs
showing values against time.  These follow the same colouring schema as the
instantaneous bar graphs.   For example, the CPU consumption line is in green
and the non-user-mode in brown.

The [history] button reveals and hides the history graphs.  The complexity of
rendering these graphics puts a noticable load on some browser systems (though
desktops are generally much more powerful than when MonDeSi was first
developed, and browser JavaScript implementations much more efficient).


PRINTING
--------
Printing a 'page' that is currently/constantly building itself (as these
displays are due to the hidden IFRAME / WebSocket asynchronously receiving the
system data updates) is problematic on all browsers tested.   As a workaround
the display updates are suspended thus stopping the page building.  The
JavaScript print is then requested.  After printing the display update is
reenabled.  Data continues to be collected during the display update
suspension.  If the [print] button remains highlighted after the print dialogue
closes click it again to resume the display updates.


CONFIGURATION
-------------
The configuration page provides a text entry area into which URLs for the nodes
to monitor are entered, one per line.  The 'nodes across window' selector allow
monitor displays to be tiled (if there is enough screen real estate).  Using
this approach widely 'separated' hosts can be monitored on a single page.  The
URL generated by the configuration dialog can be saved as a bookmark, entered
into an HTML page link, etc., for ease of subsequent access.


SUFFICIENT PRIORITY
-------------------
For a monitor such as this to be effective on a busy system it needs to have a
bit of an edge!  If INSTALLed with ALTPRI it will bump it's process priority up
to 6 when starting and reduce it back to previous when exiting (just in case
it's WASD and a persistent scripting process :-) 

  $ INSTALL REPLACE <location>MONDESI.EXE /PRIV=ALTPRI

Don't forget to reINSTALL any subsequently built or deployed executable.


CONTROLLING ACCESS
------------------
MONDESI provides an insight into the system and therefore is a security
consideration.  The information is fairly benign but for this and other reasons
it is often desirable to control who may use it.  Of course this can be
exercised using the mapping and/or authorisation functionality of the
underlying web server.  This is a WASD example that merely makes MONDESI
unknown outside of the local subnet.

  # WASD_CONFIG_MAP (HTTPD$MAP)
  if (!remote-addr:192.168.1.0/24) pass /cgi-bin/mondesi* "404 Not Found"


WEBSOCKET
---------
MONDESI v2.0ff optionally uses HTML5 WebSocket

  http://en.wikipedia.org/wiki/Websockets

to perform the data transfer from server to client.  The WASD server v10.1 and
later supports scripting using WebSockets as the IPC.  A single CGIplus
scripting process can support multiple WebSocket clients making it particularly
efficient.  The code must support multiple asynchonous threads of processing. 
MONDESI provides this using ASTs.  The WsLib..() functions from the WASD wsLIB
library handle the concurrent WebSocket processing.  See the WASD documentation
for detailed information on WebSocket support.

For multi-client WebSocket usage MonDeSi needs to be activated as a CGIplus
script.  This activation will happily support non-WebSocket clients as well. 
CGIplus activation can be ensured using a mapping rule providing redirection,
for example

  redirect /cgi-bin/mondesi ///cgiplus-bin/mondesi
  pass /mondesi/-/* /www_root/mondesi/*

MONDESI detects the underlying request type and adjusts its behaviour between
standard HTTP clients (using the "Web Push" techology described below) and
WebSocket clients.  In fact, under WASD, a single (CGIplus) process with
multiple WebSocket-connected clients can then be seconded to service a
CGI(plus) (i.e. non-WebSocket) request.  When the CGI(plus) client disconnects
WASD will run-down the process disconnecting the WebSocket clients in the
process (so-to-speak).  Not much can be done about this without having a
WebSocket specific script (perhaps wrapper).  Just an interesting wrinkle!


WEB "PUSH"
----------
MONDESI uses JavaScript and a Web 'push' technology (such as it is) to return
data (not images!) from a Web server scripting process to the client monitor
page where it is displayed as text or rendered into (crude but effective) bar
graphs.  Perhaps it would be better described as "HTTP streaming" or as "long
polling" but it *definitely* does not need to poll for each update.  The "push"
will last as long as the Web server and intermediate network connectivity
allows.

A parent page provides the JavaScript and HTML infrastructure to support
multiple monitored nodes each contained in an indepdendent, visible IFRAME,
which in turn used a dynamically created (i.e. JavaScript), hidden IFRAME to
execute the data collection URL (script).  This executes continuously,
periodically collecting data and outputting changed items into the response
stream.  Each response record contains a JavaScript tag element which contains
calls to the required data processing and display functions in the visible
IFRAME to process (only) each datum that needs updating.

If a data collection process (script) should exit for whatever reason (timeout,
network interruption, etc.) MONDESI should restart data collection
automatically.  If the monitored system become unavailable for any reason
(restart, network interruption, etc.) the monitor highlights the node
information with red after ninety seconds.  Elementary but suitable to the
desired outcome.


INTERFRAME COMMUNICATION (IFC)
------------------------------
The node(s) displayed in the parent page are not necessarily the same as the
parent page (indeed, multiple nodes could not all be) and of course these have
differing host names and/or IP addresses.  This incurs browser cross-domain
restrictions with communication between the node display HTML (i)frame and the
primary, top-level page, specifically - when needing to open/close the history
graph portion of the display associated with each node.

Two approaches for providing this IFC are used. 
Both are well documented on the Web.

1) The more elegant approach uses the modern, architected, HTML5 postMessage()
communication mechanism.  MONDESI checks for and uses this if available in the
browser.  It sends a message to the top-level page which invokes a function to
adjust the node display height.  Currently (at initial MONDESI development)
Firefox 3 (specifically Gecko 1.9), Opera 9.5 and MSIE 8 (at least) support it.

2) If postMessage() is unavailable MONDESI falls back to using an IFC *kludge*
involving the dynamic creation of an IFRAME within the node monitor IFRAME and
using this to load a URL for the served file BCHAN.HTML which contains a call
to a JavaScript function in the top-level page (to adjust the height).  The
data are contained in the hash portion of the URL and never appears outside the
browser.  This should be usable with most older, mainstream browsers.


GRAPHING
--------
This application uses a third-party JavaScript graphics toolkit to generate
both the bar and line graphs.

  Copyright (c) 2002-2009 Walter Zorn. All rights reserved.

  Performance optimizations for Internet Explorer
  by Thomas Frank and John Holdsworth.
  fillPolygon method implemented by Matthieu Haller.

  LICENSE: LGPL

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License (LGPL) as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

Used to be documented at

  http://www.walterzorn.com/jsgraphics/jsgraphics_e.htm

Now at a site maintained on behalf of the late Walter Zorn

  http://www.walterzorn.de/en/index.htm

Also accessable via WayBackMachine (split URL)

  http://replay.waybackmachine.org/
  20090302111911/http://www.walterzorn.com/jsgraphics/jsgraphics_e.htm

(and now archived in the [.JS] directory).

Many thanks to Walter for making this useful tool freely available. 
This toolkit works extraordinarily well considering it uses no graphics
capability in the browser!!  This is a non-SVG, non-VML, DOM kludge (not
meaning this as a pejorative).


BUILD DETAILS
-------------
See BUILD_MONDESI.COM procedure.

  $ @BUILD_MONDESI [LINK]


APACHE INSTALLATION
-------------------
Copy the executable to the APACHE$ROOT:[CGI-BIN] directory.

Map the JavaScript resources into web-space using (assumes the application is
located under APACHE$COMMON:[000000]):

  Alias /mondesi/-/ "/apache$common/mondesi/"


OSU ACCOMODATIONS
-----------------
OSU requires the use of the MONDESI.COM DCL wrapper to invoke the OSU CGI
environment utility before executing MONDESI.  The wrapper and the MONDESI.EXE
executable both should be placed in WWW_ROOT:[BIN].  Rather than introduce the
bulk of WASD CGILIB this deliberately slimline monitor utility checks for an
indication of the OSU environment and overtly handles the indiosyncracies.  
MONDESI does not reopen <stdout> in binary mode.  OSU requires two CGI response
headers (MONDESI supplies "Status:" and "Content-Type:").  OSU also dislikes a
response header with <CR><LF> carriage-control, so MONDESI uses only an <LF>.

Map the JavaScript resources into web-space using (assumes the application is
located under WWW_ROOT:[000000]):

  pass /mondesi/-/* /www_root/mondesi/*


WASD INSTALLATION
-----------------
Copy the executable to the CGI_EXE: directory.

Map the JavaScript resources into web-space using (assumes the application is
located in the source tree):

  # HTTPD$MAP
  pass /mondesi/-/* /wasd_root/src/mondesi/*


COPYRIGHT
---------
Copyright (C) 2008-2013 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY (update SOFTWAREVN as well!)
---------------
06-JAN-2013  MGD  v3.0.0, significant rework
                          top process monitoring
                          individual disk monitoring
                          redesign data delivery for consistency between
                            concurrent WebSocket clients
                          note the WebSocket required REDIRECT described above!
                          Happy birthday Gnomie!
08-APR-2012  MGD  v2.1.1, nudge MSIE 10.0 (Windows 8 preview) to do its best!
                          <meta http-equiv="X-UA-Compatible" ...>
28-JAN-2012  MGD  v2.1.0, migrate to WASD WebSocket wsLIB environment
                          (after significant changes to WebSocket drafts)
23-OCT-2010  MGD  v2.0.0, happy birthday Raupu
                          add network interface (NI) data
                          add processes
                          add lock waits to LOCK display
                          add SCS messages
                          PGFL data is now shown by hovering over MEM
                          data are now presented as units, kilo, mega, giga
                          support WASD WebSocket environment
                          bugfix; nodes across
03-JUN-2009  MGD  v1.0.2, bumped the dot-point merely to identify
                          the 3.05 version of wz_jsgraphics.js
13-DEC-2008  MGD  v1.0.1, CollectRMI() allow for 7.3-1 and earlier (sigh!)
                          thanks Graham Burley for suggesting sys$gl_*jobcnt
29-NOV-2008  MGD  v1.0.0, initial
*/
/*****************************************************************************/

#define SOFTWAREVN "3.0.0"
#define SOFTWARENM "MONDESI"
#ifdef __ALPHA
   char SoftwareID [] = SOFTWARENM " AXP-" SOFTWAREVN;
#endif
#ifdef __ia64
   char SoftwareID [] = SOFTWARENM " IA64-" SOFTWAREVN;
#endif
#ifdef __VAX
   char SoftwareID [] = SOFTWARENM " VAX-" SOFTWAREVN;
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unixlib.h>

/* VMS-related header files */
#include <descrip.h>
#include <dvidef.h>
#include <dvsdef.h>
#include <iodef.h>
#include <jpidef.h>
#include <libdtdef.h>
#include <lib$routines.h>
#include <lnmdef.h>
#include <psldef.h>
#include <rmidef.h>
#include <ssdef.h>
#include <starlet.h>
#include <stsdef.h>
#include <syidef.h>

#include "wslib.h"

/* required (to compile) for VMS V7.3-1 and earlier */
#ifndef RMI$_CPUCOMPAT
#define RMI$_CPUCOMPAT 21193
#define RMI$_CPUINTSTK 21194
#define RMI$_CPUMPSYNCH 21195
#define RMI$_CPUKERNEL 21196
#define RMI$_CPUEXEC 21197
#define RMI$_CPUSUPER 21198
#define RMI$_CPUUSER 21199
#define RMI$_CPUIDLE 21200
#define RMI$_PROCBATCNT 21202
#define RMI$_PROCINTCNT 21203
#define RMI$_PROCNETCNT 21204
#define RMI$_MSCP_EVERYTHING 4265
#endif

/* required for VMS V7.3-1 and earlier */
#ifdef  __VAX
globalref sys$gw_ijobcnt;
globalref sys$gw_bjobcnt;
globalref sys$gw_njobcnt;
#else
globalref sys$gl_ijobcnt;
globalref sys$gl_bjobcnt;
globalref sys$gl_njobcnt;
#endif

/* by default all structures on non-VAX are not member aligned */
#ifndef __VAX
#   pragma __nomember_alignment
#endif

#define BOOL int
#define TRUE 1
#define FALSE 0
 
#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))
 
#define FI_LI "MONDESI", __LINE__
#define EXIT_FI_LI(status) { printf ("[%s:%d]", FI_LI); exit(status); }

#define AMBIT_PRIORITY 6
#define REFRESH_INTERVAL 2

#define BAR_HEIGHT 14
#define BAR_WIDTH (GRAPH_WIDTH+2 + GRAPH_WIDTH+2 + 10)

#define NODE_HEIGHT ((BAR_HEIGHT * 9) + 20)
#define NODE_WIDTH 952

#define GRAPH_HEIGHT 100
#define GRAPH_WIDTH  300

#define PROC_HEIGHT 100
#define PROC_WIDTH  300

#define HIST_HEIGHT ((GRAPH_HEIGHT * 3) + (10 * 3))
#define HIST_HEIGHT_SCS ((GRAPH_HEIGHT * 4) + (10 * 4))
#define HIST_HEIGHT_DISK ((GRAPH_HEIGHT * 4) + (10 * 4))

/* exit after fifteen minutes without a WebSocket connection */
#define EXIT_AFTER_SECONDS 15 * 60;

/* maximum number of network devices */
#define NI_DEVICE_MAX 8

/* maximum number of disks that can be monitored */
#define DISK_DEVICE_MAX 32
#define DISK_NAME_SIZE 16
#define DISK_VOLNAM_SIZE 12
/* maximim number of disks readable from system */
#define DISK_OFFSETS_MAX 100

/* maximum number of processes that can be reported */
#define PROC_MAX 5
#define PROC_NAME_SIZE 15
/* maximum number of processes readable from system */
#define PROC_CLASS_MAX 200

/*
Proportional fonts derived from:
http://www.codestyle.org/css/font-family/sampler-SansSerif.shtml
*/
#define FONT_PROPORTIONAL "\"Lucida Sans\",Arial,Helvetica,Tahoma,sans-serif"
#define FONT_PROP_SIZE 12

/*
Monospace fonts derived from:
http://www.codestyle.org/css/font-family/sampler-Monospace.shtml
*/
#define FONT_MONOSPACE "\"Lucida Sans Typewriter\",\"Lucida Console\",\
\"Andale Mono\",\"Consolas\",\"Monaco\",\"DejaVu Sans Mono\",Courier,\
\"Courier New\",monospace,serif" 
#define FONT_MONO_PROC_SIZE 7

/* font size in points for control buttons */
#define FONT_BTN_SIZE 7

/* font size in points for notes */
#define FONT_NOTES_SIZE 9

/* font size in points for control buttons */
#define FONT_BTN_SIZE 7

/*******************/
/* data structures */
/*******************/

typedef struct STRUCT_IOSBLK
{
   unsigned short  iosb$w_status;
   unsigned short  iosb$w_bcnt;
   unsigned int    iosb$l_reserved;
}
IOSBLK;

/* from SPIDEF.H REMIDEF.H */
#ifndef __VAX
   /* these members are aligned on non-VAX */
#  pragma __member_alignment __save 
#  pragma __member_alignment 
#  ifdef __INITIAL_POINTER_SIZE
      /* and 32 bit pointers */
#     pragma __required_pointer_size __save
#     pragma __required_pointer_size __short
#  endif
#endif

typedef struct STRUCT_SYSTEM_DATA
{
   unsigned long  ActiveCpuCount,
                  AvailCpuCount,
                  BufferedIO,
                  ClusterMsgSent,
                  ClusterMsgRcvd,
                  ClusterKbyteMap,
                  ClusterKbyteSent,
                  ClusterKbyteRcvd,
                  ClusterMember,
                  ClusterEVotes,
                  ClusterNodes,
                  ClusterQuorum,
                  ClusterVotes,
                  CpuCount,
                  DirectIO,
                  FcpRead,
                  FcpWrite,
                  FreeList,
                  HardPageFaults,
                  LockActivity,
                  LockWait,
                  MemInUse,
                  MemInUseMax,
                  MemSize,
                  ModList,
                  MscpRead,
                  MscpWrite,
                  PageFaults,
                  PageFaultsHard,
                  PageFaultsSoft,
                  PageFileFree,
                  PageFilePage,
                  PageFilePercent,
                  PageReadIO,
                  PageSize,
                  PageWriteIO,
                  PrevBufferedIO,
                  PrevBusyCpu,
                  PrevClusterMsgSent,
                  PrevClusterMsgRcvd,
                  PrevClusterKbyteMap,
                  PrevClusterKbyteSent,
                  PrevClusterKbyteRcvd,
                  PrevClusterNodes,
                  PrevWebSockConnectedCount,
                  PrevDirectIO,
                  PrevFcpRead,
                  PrevFcpWrite,
                  PrevHardPageFaults,
                  PrevLockActivity,
                  PrevLockWait,
                  PrevMemInUse,
                  PrevMountCount,
                  PrevMscpRead,
                  PrevMscpWrite,
                  PrevPageFaults,
                  PrevPageFilePercent,
                  PrevProcessCount,
                  PrevProcessIntCount,
                  PrevUserCpu,
                  ProcessCount,
                  ProcessBatCount,
                  ProcessIntCount,
                  ProcessNetCount,
                  Processes;
 
#ifndef __VAX
   unsigned __int64  NetIntCount,
                     NetIntBlocksRx,
                     NetIntBlocksTx,
                     NetIntBytesRx,
                     NetIntBytesTx,
                     NetIntErrorsHard,
                     NetIntErrorsSoft;
#else
   unsigned long  NetIntCount,
                  NetIntBlocksRx,
                  NetIntBlocksTx,
                  NetIntBytesRx,
                  NetIntBytesTx,
                  NetIntErrorsHard,
                  NetIntErrorsSoft;
#endif

#ifndef __VAX
   unsigned __int64  PrevNetIntBlocksRx,
                     PrevNetIntBlocksTx,
                     PrevNetIntBytesRx,
                     PrevNetIntBytesTx;
#else
   int  PrevNetIntBlocksRx,
        PrevNetIntBlocksTx,
        PrevNetIntBytesRx,
        PrevNetIntBytesTx;
#endif

   char  ArchName [15+1],
         HwName [60+1],
         NodeName [15+1],
         OutputBuffer [2048+(64*DISK_DEVICE_MAX)],
         ProcessBuffer [256+(48*6*PROC_MAX)],
         VmsVersion [8+1];

   unsigned long  BootBinTime [2],
                  CpuTicksBusy [2],
                  CpuTicksUser [2];

   int  CpuBusyPercent,
        CpuUserPercent,
        MonValLength,
        NodeInfoLength,
        NodeInfoTimeStamp,
        PrevInit,
        PrevMinute,
        ProcessBufferLength,
        ProcMonitor;

   int  DiskCount,
        MountCount;
   char  DiskName [DISK_DEVICE_MAX][DISK_NAME_SIZE+1],
         DiskVolName [DISK_DEVICE_MAX][DISK_VOLNAM_SIZE+1];
   unsigned long  DiskOptCnt [DISK_DEVICE_MAX],
                  DiskOperCount [DISK_DEVICE_MAX],
                  DiskPrevOperCount [DISK_DEVICE_MAX],
                  DiskPrevOptCnt [DISK_DEVICE_MAX];

   int  ProcCount,
        ProcSanityCheck;
   char  ProcName [PROC_MAX][PROC_NAME_SIZE+1];
   unsigned long  ProcBioCnt [PROC_MAX],
                  ProcCpuTim [PROC_MAX],
                  ProcDioCnt [PROC_MAX],
                  ProcEpid [PROC_MAX],
                  ProcPageFlts [PROC_MAX];
   int (*ProcSpiFunction)(struct STRUCT_SYSTEM_DATA*);
}
SYSTEM_DATA;

/* used for multiple per-WebSocket client and CGI client (for convenience) */
typedef struct STRUCT_CLIENT_DATA
{
   int  NodeInfoTimeStamp,
        ProcMonitor;

   char  InputBuffer [1024];

   struct  WsLibStruct  *WsLibPtr;

} CLIENT_DATA;

#ifndef __VAX
#   pragma __member_alignment __restore
#endif

/******************/
/* global storage */
/******************/

BOOL  Debug,
      ApacheHttpd,
      OsuHttpd,
      UsingCGI,
      WasdHttpd;

int  CgiNodeInfoTimeStamp,
     IsCgiPlus,
     PrevPrio,
     RefreshInterval,
     UsageCount,
     WebSockConnectedCount;

unsigned int  CollectDataSeconds,
              ExitTimeStamp,
              SpiEfn,
              TimeStamp;

unsigned long  CurrentBinTime [2];
unsigned short  CurrentNumTime [7];

float  DeltaSeconds = 999999.0;

char  *CgiHttpRefererPtr,
      *CgiHttpUserAgentPtr,
      *CgiScriptNamePtr,
      *CgiQueryStringPtr,
      *CgiPlusEofPtr = NULL,
      *CgiPlusEotPtr = NULL,
      *CgiPlusEscPtr = NULL;

char  NetIntDev [NI_DEVICE_MAX*24],
      NotePad [128],
      RefererHost [128];

SYSTEM_DATA  SystemData;

/**************/
/* prototypes */
/**************/

void CollectData ();
int CollectDiskSPI (SYSTEM_DATA*);
int CollectNetInt (SYSTEM_DATA*);
int CollectProcSPI1 (SYSTEM_DATA*);
int CollectProcSPI2 (SYSTEM_DATA*);
int CollectRMI (SYSTEM_DATA*);
int CollectSyiStatic (SYSTEM_DATA*);
int CollectSyiDynamic (SYSTEM_DATA*);

void MonDeSiForm (char*, int);
void MonDeSiFormatData (SYSTEM_DATA*);
void MonDeSiHtmlData ();
void MonDeSiNode (char*, SYSTEM_DATA*);
void MonDeSiPage (char*, int);
void ReturnPrio ();
int UrlDecode (char*);
char *MungeProcName (char*);

int exe$getspi (__unknown_params);
int sys$getrmi (__unknown_params);

void WebSockDestroy (struct WsLibStruct*);
void WebSockReadAst (struct WsLibStruct*);
void WebSockRequest ();

/*****************************************************************************/
/*
*/

int main ()

{
   static unsigned short  LogValueLength;
   static char  LogValue [32];
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LogValue)-1, LNM$_STRING, &LogValue, &LogValueLength },
      { 0,0,0,0 }
   };
   static $DESCRIPTOR (LogNameDsc, "MONDESI_REFRESH");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   int  status,
        NodesAcross;
   char  *cptr, *sptr, *zptr,
         *NodeList;
   char  IframeId [64];

   /*********/
   /* begin */
   /*********/

   /* ambit attempt to bump the process priority (if INSTALLed with ALTPRI) */
   atexit (&ReturnPrio);
   sys$setpri (0, 0, AMBIT_PRIORITY, &PrevPrio, 0, 0);

   if (getenv ("MONDESI$DBUG")) Debug = 1;

   if (Debug)
      fprintf (stdout,
"Status: 200\n\
Content-Type: text/plain\n\
Script-Control: X-content-encoding-gzip=0\n\
\r\n");

   ApacheHttpd = (getenv ("APACHE$SHARED_SOCKET") != NULL ||
                  getenv ("APACHE$PARENT_PID") != NULL ||
                  getenv ("APACHE$SERVER_TAG") != NULL);
   OsuHttpd = (getenv("WWWEXEC_RUNDOWN_STRING") != NULL);
   WasdHttpd = (getenv("HTTP$INPUT") != NULL);

   if (!Debug && !OsuHttpd)
      if (!(stdout = freopen ("SYS$OUTPUT:", "w", stdout, "ctx=bin")))
         EXIT_FI_LI (vaxc$errno);                                           

   IsCgiPlus = ((CgiPlusEofPtr = getenv("CGIPLUSEOF")) != NULL);
   if ((CgiPlusEscPtr = getenv("CGIPLUSESC")) == NULL) CgiPlusEscPtr = "";
   if ((CgiPlusEotPtr = getenv("CGIPLUSEOT")) == NULL) CgiPlusEotPtr = "";

   status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
   if (VMSnok (status))
      LogValue[0] = '\0';
   else
      LogValue[LogValueLength] = '\0';
   RefreshInterval = atoi(LogValue);
   if (RefreshInterval <= 0 || RefreshInterval > 60)
      RefreshInterval = REFRESH_INTERVAL;

   memset (&SystemData, 0, sizeof(SystemData));
   CollectSyiStatic (&SystemData);
   SystemData.NetIntCount = CollectNetInt (NULL);
   SystemData.DiskCount = CollectDiskSPI (&SystemData);
   SystemData.ProcCount = CollectProcSPI1 (NULL);
   SystemData.ProcSpiFunction = &CollectProcSPI1;
   CollectData ();

   do {

      /* with CGIplus this call will block waiting for the next request */
      WsLibCgiVar ("");

      /********************/
      /* request received */
      /********************/

      UsageCount++;

      CgiScriptNamePtr = WsLibCgiVar("SCRIPT_NAME");
      CgiQueryStringPtr = WsLibCgiVar("QUERY_STRING");
      CgiHttpRefererPtr = WsLibCgiVar("HTTP_REFERER");
      CgiHttpUserAgentPtr = WsLibCgiVar("HTTP_USER_AGENT");

      /* derive the URL scheme and host from the HTTP referrer field */
      zptr = (sptr = RefererHost) + sizeof(RefererHost)-1;
      cptr = CgiHttpRefererPtr;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';

      if (isdigit(*CgiQueryStringPtr))
      {
         /********/
         /* data */
         /********/

         if (!SpiEfn)
            if (VMSnok (status = lib$get_ef (&SpiEfn)))
               EXIT_FI_LI (status);

         UsingCGI = (WsLibCgiVarNull ("WEBSOCKET_INPUT") == NULL);

         if (UsingCGI)
            MonDeSiHtmlData ();
         else
         {
            if (!WsLibIsCgiPlus())
            {
               /* WebSocket must be via a persistent process */
               fprintf (stdout,
"Status: 403 Forbidden\r\n\
\r\n\
MonDeSi (using WebSocket) must be accessed via a CGIplus mapping!");
               fflush (stdout);
            }
            else
               WebSockRequest ();
         }
      }
      else
      {
         /********/
         /* page */
         /********/

         NodesAcross = 1;
         if (cptr = strstr (CgiQueryStringPtr, "across="))
            NodesAcross = atoi(cptr+7);
         if (NodesAcross < 0 || NodesAcross > 4) NodesAcross = 1;

         IframeId[0] = '\0';
         if (cptr = strstr (CgiQueryStringPtr, "iframe="))
         {
            zptr = (sptr = IframeId) + sizeof(IframeId)-1;
            for (cptr += 7;
                 *cptr && *cptr != '&' && sptr < zptr;
                 *sptr++ = *cptr++);
            *sptr = '\0';
         }

         if (NodeList = strstr (CgiQueryStringPtr, "nodes="))
         {
            for (sptr = (NodeList += 6); *sptr && *sptr != '&'; sptr++);
            *sptr = '\0';
            UrlDecode (NodeList);
            /* eliminate those pesky W*s <CR><LF>s */
            for (cptr = sptr = NodeList; *cptr; *sptr++ = *cptr++)
               if (*(unsigned short*)cptr == '\r\n') cptr++;
            *sptr = '\0';
            /* compress white-space */
            for (cptr = sptr = NodeList; *cptr; *sptr++ = *cptr++)
               while (isspace(*cptr) && isspace(*(cptr+1))) cptr++;
            *sptr = '\0';
         }
         else
            NodeList = "";

         if (ispunct(*CgiQueryStringPtr))
            MonDeSiForm (NodeList, NodesAcross);
         else
         if (IframeId[0])
            MonDeSiNode (IframeId, &SystemData);
         else
         if (*NodeList)
            MonDeSiPage (NodeList, NodesAcross);
         else
            MonDeSiPage (SystemData.NodeName, 1);
      }

      if (IsCgiPlus)
      {
         /* send CGIplus end-of-request sentinal */
         fflush (stdout);
         fputs (CgiPlusEofPtr, stdout);
         fflush (stdout);
      }

   } while (IsCgiPlus);

   exit (SS$_NORMAL);
} 

/*****************************************************************************/
/*
Return to previous process priority (in case ambit raise was successful).
*/

void ReturnPrio ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (!PrevPrio) return;

   status = sys$setpri (0, 0, PrevPrio, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$setpri() %%X%08.08X\n", status);
} 

/*****************************************************************************/
/*
Provide the client with the configuration setup form.
*/

void MonDeSiForm
(
char *NodeList,
int NodesAcross
)
{
   char  *cptr, *sptr;
   char  NodeUrl [256];

   /*********/
   /* begin */
   /*********/

   if (!*NodeList)
   {
      /* concoct a node list entry */
      sptr = WsLibCgiVar ("REQUEST_SCHEME");
      if (!*sptr)
      {
         sptr = WsLibCgiVar ("SERVER_PORT");
         if (atoi(sptr) == 443)
            sptr = "https:";
         else
            sptr = "http:";
      }
      cptr = WsLibCgiVar ("HTTP_HOST");
      sprintf (NodeList = NodeUrl, "%s//%s%s", sptr, cptr, CgiScriptNamePtr);
   }

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<html>\n\
<head>\n\
<meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n\
\
<style type=\"text/css\">\n\
\
body { background-color:DarkSlateGray; margin:1em; \
font-family:%s; font-size: %dpt; text-align:center; }\n\
input { border-width:1px; border-style:outset; }\n\
textarea,select { border-width:1px; border-style:inset; }\n\
.pageTitle { font-size:140%%; font-weight:bold; letter-spacing:1px; \
text-decoration:underline; color:white; }\n\
.subTitle { font-size:90%%; font-style:italic; letter-spacing:1px; \
color:white; }\n\
\
</style>\n\
\
<title>%s</title>\n\
</head>\n\
<body>\n\
<form method=\"GET\" action=\"%s\" \
enctype=\"application/x-www-form-urlencoded\" autocomplete=\"off\">\n\
\
<span class=\"pageTitle\">Monitor de Sistema</span>\n\
<p style=\"margin-top:0.5em;\">\
<span class=\"subTitle\">~ System Monitor ~</span>\n\
\
<p><textarea id=\"nodes\" name=\"nodes\" rows=\"10\" cols=\"64\" \
scroll=\"auto\">\
%s</textarea>\n\
\
<p><select name=\"across\">\n\
<option value=\"1\"%s>1\n\
<option value=\"2\"%s>2\n\
<option value=\"3\"%s>3\n\
<option value=\"4\"%s>4\n\
</select>\
<span style=\"color:white;\">&nbsp;&nbsp;node(s) across window</span>\n\
\
<script language=\"JavaScript\"> \
if (\"WebSocket\" in window) { \
document.write(\'<p><input type=\"checkbox\" name=\"nows\" value=\"1\"%s>\
&nbsp;<span style=\"color:white;\">ignore WebSocket</span>\'); } \
</script>\n\
<p><input type=\"submit\" value=\" Monitor \">&nbsp;&nbsp;\n\
<input type=\"reset\" value=\" Reset \" \
onclick=\"return confirm(\'Are you sure?\')\">&nbsp;&nbsp;\n\
<input type=\"reset\" value=\" Clear \" \
onclick=\"if (confirm(\'Are you sure?\')) \
document.getElementById(\'nodes\').value=\'\'; return false;\">\n\
\
</form>\n\
</body>\n\
</html>\n",
            FONT_PROPORTIONAL, FONT_PROP_SIZE,
            SoftwareID,
            CgiScriptNamePtr,
            /* should be no HTML-forbidden characters in this! */
            NodeList,
            NodesAcross == 1 ? " selected" : "",
            NodesAcross == 2 ? " selected" : "",
            NodesAcross == 3 ? " selected" : "",
            NodesAcross == 4 ? " selected" : "",
            atoi(WsLibCgiVar("FORM_NOWS")) ? " checked" : "");
} 

/*****************************************************************************/
/*
Provide the page that contains one or more monitor panels (IFRAMEs containing
their own HTML 'page' content).
*/

void MonDeSiPage
(
char *NodeList,
int NodesAcross
)
{
   int  NodeCount = 0;
   char  *aptr, *cptr, *sptr, *zptr;
   char  HostUrl [128];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiPage() |%s|\n", NodeList);

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<html>\n\
<head>\n\
<meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n\
\
<style type=\"text/css\">\n\
\
body { background-color:DarkSlateGray; margin:1em; \
font-family:%s; font-size:%dpt; text-align:center; }\n\
@media print { body { background-color:white; } }\n\
input { border-width:1px; border-style:outset; }\n\
@media print { input { visibility:hidden; } }\n\
.node_iframe { height:%dpx; width:%dpx; border:0; padding:0; margin:5px; }\n\
.nodes_across { border:0; padding:0; margin:0; }\n\
.nodes_across td { vertical-align:top; }\n\
\
</style>\n\
\
<script language=\"JavaScript\">\n\
\
function mdsMsgChannel (msg)\n\
{\n\
   // message posting browsers (HTML5)\n\
   if (msg.data.substring(0,3) == \"mds\")\n\
   {\n\
      eval(msg.data);\n\
      return;\n\
   }\n\
   if (msg.data != \"\") alert(\"ERROR:\"+msg.data);\n\
}\n\
\
try { var mdsMessageListener = \
window.addEventListener(\"message\",mdsMsgChannel,false); } \
catch(err) { }\n\
\
function mdsBackChannel (forEval)\n\
{\n\
   // inter-frame kludge\n\
   if (forEval.substring(0,4) == \"#mds\")\n\
   {\n\
      eval(forEval.substr(1));\n\
      return;\n\
   }\n\
   if (forEval != \"\") alert(\"ERROR:\"+forEval.substr(1));\n\
}\n\
\
function mdsIframeHeight (iframeId,toTop,height)\n\
{\n\
   // remotely called via IFC\n\
   var oNodeFrame = document.getElementById(iframeId);\n\
   oNodeFrame.height = oNodeFrame.style.height = height + \"px\";\n\
   if (toTop) try { if (oNodeFrame.offsetTop > 100) \
window.scrollTo(0,oNodeFrame.offsetTop-10); } catch (err) { }\n\
}\n\
\
</script>\n\
<title>%s</title>\n\
</head>\n\
<body>\n",
            FONT_PROPORTIONAL, FONT_PROP_SIZE,
            NODE_HEIGHT, NODE_WIDTH,
            SoftwareID);

   cptr = NodeList;
   while (*cptr)
   {
      while (*cptr && isspace(*cptr)) cptr++;
      zptr = (sptr = HostUrl) + sizeof(HostUrl)-1;
      /* parse to the end of the scheme */
      while (*cptr && !isspace(*cptr) && *cptr != '/' && sptr < zptr)
         *sptr++ = *cptr++;
      if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      /* if no domain name */
      if (!strchr (HostUrl, '.'))
      {
         /* same host, just use the script component */
         sptr = HostUrl;
         for (aptr = CgiScriptNamePtr; *aptr && sptr < zptr; *sptr++ = *aptr++);
      }
      else
      /* if provides a script component */
      if (*cptr && !isspace(*cptr))
      {
         if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
         while (*cptr && !isspace(*cptr) && sptr < zptr) *sptr++ = *cptr++;
      }
      else
      {
         /* append a script component */
         for (aptr = CgiScriptNamePtr; *aptr && sptr < zptr; *sptr++ = *aptr++);
      }
      *sptr = '\0';
      while (*cptr && !isspace(*cptr)) cptr++;
      while (*cptr && isspace(*cptr)) cptr++;

      if (NodesAcross > 1)
      {
         if (!(NodeCount % NodesAcross))
         {
            if (!NodeCount)
               fputs ("<table class=\"nodes_across\">\n<tr><td>\n", stdout);
            else
               fputs ("</td></tr>\n<tr><td>\n", stdout);
         }
         else
            fputs ("</td><td>\n", stdout);
      }
      NodeCount++,

      fprintf (stdout,
"<iframe id=\"node%d\" class=\"node_iframe\" \
frameborder=\"0\" scrolling=\"no\" \
src=\"%s?iframe=node%d&nows=%d\"></iframe>\n",
               NodeCount,
               HostUrl,
               NodeCount,
               atoi(WsLibCgiVar("FORM_NOWS")) ? 1 : 0);
   }

   if (NodesAcross > 1) fputs ("</td></tr>\n</table>\n", stdout);

   fprintf (stdout,
"<form method=\"GET\" action=\"%s\" \
enctype=\"application/x-www-form-urlencoded\" autocomplete=\"off\">\n\
<input type=\"hidden\" name=\"!\" value=\"\">\n\
<input type=\"hidden\" name=\"nows\" value=\"%d\">\n\
<input type=\"hidden\" name=\"nodes\" value=\"%s\">\n\
<input type=\"hidden\" name=\"across\" value=\"%d\">\n\
<input type=\"submit\" value=\" Configure \">\n\
</form>\n\
</body>\n\
</html>\n",
            CgiScriptNamePtr,
            atoi(WsLibCgiVar("FORM_NOWS")) ? 1 : 0,
            strchr(NodeList,'.') ? NodeList : "",
            NodesAcross);
} 

/*****************************************************************************/
/*
Provide the in-frame node display.
*/

void MonDeSiNode
(
char *IframeId,
SYSTEM_DATA *sysptr
)
{
   int  cnt,
        WebSocketOk = 0;
   char  *cptr;
   char  DivStyle [256];

   /*********/
   /* begin */
   /*********/

   /* if WASD and the version >= 10.1 then WebSocket can be used */
   if (!atoi(WsLibCgiVar("FORM_NOWS")))
      if (cptr = WsLibCgiVarNull("SERVER_SOFTWARE"))
         if (cptr = strstr (cptr, "WASD/")) /* e.g. "WASD/10.2" */
            if (WebSocketOk = (((atoi(cptr+5) >= 10) &&
                                (atoi(cptr+8) >= 1)) ||
                               (atoi(cptr+5) >= 11)))
               if (cptr = getenv ("MONDESI_WEBSOCKET"))
                  WebSocketOk = atoi(cptr);

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<html>\n\
<head>\n\
<meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n\
\
<style type=\"text/css\">\n\
\
body { background-color:white; margin:0; padding:0; \
font-family:%s; font-size:%dpt; }\n\
.node_block { width:100%%; background-color:beige; color:black; \
border-style:outset; border-width:1px; border-color:black; \
margin:0; padding:0; text-align:left; }\n\
@media print { .node_block { background-color:white; \
border-style:solid; border-width:2px; } }\n\
.node_block td { margin:10px; padding:0; }\n\
.node_control { width:%dpx; height:18px; margin:0 0 0 5px; \
white-space:nowrap; }\n\
.node_graph { border-width:0; margin:0; padding:0; }\n\
.node_graph td { margin:0; padding:0; }\n\
.node_notes_td { width:%dpx; vertical-align:top; }\n\
.node_notes { font-size:%dpt; line-height:12px; padding:5px; }\n\
.node_name { font-size:120%%; font-weight:bold; line-height:15px; \
text-decoration:underline; display:block; margin:0 0 3px 0; }\n\
.node_time { font-size:90%%; line-height:15px; display:block; }\n\
.node_text { font-size:80%%; line-height:14px; display:inline; }\n\
.proc_wrapt { display:none; width:100%%; height:%dpx; table-layout:fixed; \
font-family:%s; font-size:%dpt; }\n\
.proc_wrapt td { vertical-align:top; width:100%%; }\n\
.proc_table { width:100%%; table-layout:fixed; \
font-family:%s; font-size:100%%; padding:0 1em 1em 1em; }\n\
.proc_table th { width:19%%; text-align:left; white-space:pre; \
text-decoration:underline; padding:0; margin:1px; }\n\
.proc_table td { width:19%%; text-align:left; white-space:pre; \
padding:0; padding-right:1em; margin:1px; }\n\
.history_graph { border-width:0; margin:0; padding:0 0 5px 0; }\n\
.history_graph td { margin:0; padding:0; }\n\
.btn { font-size:%dpt; border-style:solid; border-width:1px; \
border-color:gray; height:18px; padding:1px; margin:0 3px 0 0; }\n\
@media print { .node_control { visibility:hidden; } }\n\
\
</style>\n\
<script language=\"JavaScript\">\n\
\
var WASDwebSocket = %s;\n\
var mdsBarWidth = %d;\n\
var mdsBarHeight = %d;\n\
var mdsBody;\n\
var mdsCpuCount = %d;\n\
var mdsCluster = %s;\n\
var mdsDiskCount = %d;\n\
var mdsDiskMount = 0;\n\
var mdsExeVersion = \"%s\";\n\
var mdsGraphWidth = %d;\n\
var mdsGraphHeight = %d;\n\
var mdsHistVisible = false;\n\
var mdsIframeId = \"%s\";\n\
var mdsNodeName = \"%s\";\n\
var mdsProcCount = %d;\n\
var mdsProcHeight = %d;\n\
var mdsRefreshInterval = %d;\n\
var mdsRefererHost = \"%s\"\n\
var mdsScriptName = \"%s\";\n\
\
</script>\n\
<script language=\"JavaScript\" src=\"%s%s\"></script>\n\
<script language=\"JavaScript\" src=\"%s%s\"></script>\n\
\
<title>%s</title>\n\
</head>\n\
<body>\n",
            FONT_PROPORTIONAL, FONT_PROP_SIZE,
            GRAPH_WIDTH+2,
            GRAPH_WIDTH+2,
            FONT_NOTES_SIZE,
            PROC_HEIGHT,
            FONT_MONOSPACE, FONT_MONO_PROC_SIZE,
            FONT_MONOSPACE,
            FONT_BTN_SIZE,
            WebSocketOk ? "true" : "false",
            BAR_WIDTH, BAR_HEIGHT,
            SystemData.ActiveCpuCount,
            SystemData.ClusterMember ? "true" : "false",
            SystemData.DiskCount,
            SOFTWAREVN,
            GRAPH_WIDTH, GRAPH_HEIGHT,
            IframeId,
            SystemData.NodeName,
            SystemData.ProcCount ? 5 : 0,
            PROC_HEIGHT,
            RefreshInterval,
            RefererHost,
            CgiScriptNamePtr,
            RefererHost, "/mondesi/-/js/mondesi.js",
            RefererHost, "/mondesi/-/js/wz_jsgraphics.js",
            SoftwareID);

   fprintf (stdout, "<table class=\"node_block\">\n<tr><td>\n");

   /**************/
   /* bar graphs */
   /**************/

   sprintf (DivStyle,
"position:relative;height:%dpx;width:%dpx;margin-left:5px;",
            BAR_HEIGHT, BAR_WIDTH);

   fprintf (stdout,
"<table id=\"node_graph\" class=\"node_graph\">\n\
<tr><td class=\"node_notes_td\">\
<div id=\"node_notes\" class=\"node_notes\"></div></td>\n\
\
<td style=\"vertical-align:top;\" rowspan=\"2\">\n\
<div id=\"node_cpu\" style=\"%smargin-top:5px;\"></div>\n\
<div id=\"node_mem\" style=\"%s\"></div>\n\
<div id=\"node_pgfl\" style=\"%sdisplay:none;\"></div>\n\
<div id=\"node_flts\" style=\"%s\"></div>\n\
<div id=\"node_bio\" style=\"%s\"></div>\n\
<div id=\"node_dio\" style=\"%s\"></div>\n\
<div id=\"node_lock\" style=\"%s\"></div>\n\
<div id=\"node_ni\" style=\"%s\"></div><!-- %s -->\n\
<div id=\"node_disk\" style=\"%s\"></div>\n\
<div id=\"node_prc\" style=\"%s\"></div>\n",
            DivStyle, DivStyle, DivStyle, DivStyle,
            DivStyle, DivStyle, DivStyle,
            DivStyle, NetIntDev, DivStyle, DivStyle);

   if (sysptr->ClusterMember)
      fprintf (stdout,
"<div id=\"node_scs\" style=\"display:none;\">\n\
<div id=\"node_scsmg\" style=\"%s\"></div>\n\
<div id=\"node_scskb\" style=\"%s\"></div>\n\
<div id=\"node_mscp\" style=\"%s\"></div>\n\
</div>\n",
               DivStyle, DivStyle, DivStyle);

   if (sysptr->DiskCount)
   {
      fprintf (stdout,
"<div id=\"node_disk_device\" style=\"display:none;\">\n");

      for (cnt = 0; cnt < sysptr->DiskCount; cnt++)
         fprintf (stdout, "<div id=\"node_disk_%d\" style=\"%s\"></div>\n",
                  cnt, DivStyle);

      fprintf (stdout, "</div>\n");
   }

   fprintf (stdout,
"</td></tr>\n\
\
<tr><td style=\"height:100%%;vertical-align:bottom;padding:0 0 5px 0;\">\
<div id=\"node_control\" class=\"node_control\">\n\
<input type=\"hidden\" name=\"refresh\" value=\"%d\">\n\
<input id=\"history_btn\" class=\"btn\" type=\"submit\" value=\"history\" \
onclick=\"mdsHistoryClick()\" style=\"width:40px;\">\
<input id=\"scs_btn\" class=\"btn\" type=\"submit\" value=\"cluster\" \
onclick=\"mdsClusterClick()\" style=\"width:38px;\" disabled>\
<input id=\"disk_btn\" class=\"btn\" type=\"submit\" value=\"disk\" \
onclick=\"mdsDiskClick()\" style=\"width:32px;\" disabled>\
<input id=\"proc_btn\" class=\"btn\" type=\"submit\" value=\"proc\" \
onclick=\"mdsProcClick()\" style=\"width:32px;\" disabled>\
<input type=\"submit\" class=\"btn\" value=\"scale\" \
onclick=\"return mdsScale();\" style=\"width:32px;\">\
<input id=\"zero_btn\" class=\"btn\" type=\"submit\" value=\"zero\" \
onclick=\"return mdsNodeZero();\" style=\"width:32px;\">\
<input id=\"print_btn\" class=\"btn\" type=\"submit\" value=\"print\" \
onclick=\"mdsPrintClick();return false;\" style=\"width:32px;\">\
<input id=\"pause_btn\" class=\"btn\" type=\"submit\" value=\"pause\" \
onclick=\"return mdsPauseClick();\" style=\"width:42px;\">\n\
</div></td></tr>\n\
\
</table>\n\
</td></tr>\n\
<tr><td style=\"padding-top:5px;\">\n",
             RefreshInterval);

   /******************/
   /* history graphs */
   /******************/

   sprintf (DivStyle,
"position:relative;height:%dpx;width:%dpx;margin:0 5px 7px 5px;",
            GRAPH_HEIGHT+2, GRAPH_WIDTH+2);

   /* MSIE<10 table.innerHTML is read-only so wrap with a div */
   fprintf (stdout,
"<table id=\"proc_wrapt\" class=\"proc_wrapt\">\n\
<tr><td><div id=\"proc_table_div\"></div></td></tr>\n\
</table>\n\
<table id=\"history_graph\" class=\"history_graph\">\n\
<tr>\n\
<td><div id=\"hist_cpu\" style=\"%s\"></div></td>\n\
<td><div id=\"hist_mem\" style=\"%s\" \
onmouseover=\"mdsForeFront(\'hist_pgfl\',\'hist_mem\')\"></div>\
<div id=\"hist_pgfl\" style=\"%sdisplay:none;\" \
onmouseout=\"mdsForeFront(\'hist_mem\',\'hist_pgfl\')\"></div></td>\n\
<td><div id=\"hist_flts\" style=\"%s\"></div></td>\n\
</tr>\n\
\
<tr>\n\
<td><div id=\"hist_bio\" style=\"%s\"></div></td>\n\
<td><div id=\"hist_dio\" style=\"%s\"></div></td>\n\
<td><div id=\"hist_lock\" style=\"%s\"></div></td>\n\
</tr>\n\
\
<tr>\n\
<td><div id=\"hist_ni\" style=\"%s\"></div></td>\n\
<td><div id=\"hist_disk\" style=\"%s\"></div></td>\n\
<td><div id=\"hist_prc\" style=\"%s\"></div></td>\n\
</tr>\n",
            DivStyle, DivStyle, DivStyle, DivStyle,
            DivStyle, DivStyle, DivStyle,
            DivStyle, DivStyle, DivStyle);

   if (sysptr->ClusterMember)
      fprintf (stdout,
"<tr id=\"hist_scs\" style=\"display:none;\">\n\
<td><div id=\"hist_scsmg\" style=\"%s\"></div></td>\n\
<td><div id=\"hist_scskb\" style=\"%s\"></div></td>\n\
<td><div id=\"hist_mscp\" style=\"%s\"></div></td>\n\
</tr>\n",        
               DivStyle, DivStyle, DivStyle);

   if (sysptr->DiskCount)
   {
      for (cnt = 0; cnt < sysptr->DiskCount; cnt++)
      {
         if (cnt && cnt % 3 == 0) fprintf (stdout, "</tr>\n");
         if (cnt % 3 == 0) fprintf (stdout, "<tr>\n");
         fprintf (stdout,
"<td><div id=\"hist_disk_%d\" style=\"%s\"></div></td>\n",
                  cnt, DivStyle);
      }

      if (cnt % 3 != 3) fprintf (stdout, "</tr>\n");
   }

   fputs ("</table>\n</td></tr>\n</table>\n", stdout);

   /**************/
   /* javascript */
   /**************/

   fprintf (stdout,
"<script language=\"JavaScript\">\n\
\
mdsBody = document.getElementsByTagName(\"body\").item(0);\n\
// give CGIplus (WebSocket) a chance to complete\n\
setTimeout(\"mdsInitialStart();\",500);\n\
mdsNodeNotes(\"<span class=\'node_name\'>%s</span>\");\n\
\
mdsNodeInit(\"cpu\",true);\n\
mdsNodeInit(\"mem\",true);\n\
mdsNodeInit(\"pgfl\",true);\n\
mdsNodeInit(\"flts\",true);\n\
mdsNodeInit(\"bio\",true);\n\
mdsNodeInit(\"dio\",true);\n\
mdsNodeInit(\"lock\",true);\n\
mdsNodeInit(\"ni\",%s);\n\
mdsNodeInit(\"disk\",true);\n\
mdsNodeInit(\"prc\",true);\n",
            SystemData.NodeName,
            SystemData.NetIntCount ? "true" : "false");

   if (sysptr->ClusterMember)
      fprintf (stdout,
"mdsNodeInit(\"scsmg\",true);\n\
mdsNodeInit(\"scskb\",true);\n\
mdsNodeInit(\"mscp\",true);\n");

   if (sysptr->DiskCount)
      for (cnt = 0; cnt < sysptr->DiskCount; cnt++)
         fprintf (stdout, "mdsNodeInit(\"disk_%d\",true);\n", cnt);

   fputs ("</script>\n</body>\n</html>\n", stdout);
} 

/*****************************************************************************/
/*
For an HTTP request.  At the defined interval gather dynamic system and
performance data and for appropriate (changed) data call the relevant
JavaScript (RPC) function in the parent panel (window) to process and graph
associated data.  Do the best we can to disable proxy caching ensuring a timely
flow of response data.
*/

void MonDeSiHtmlData ()

{
   int  status,
        HostUrlLength,
        TimeStamp;
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiHtmlData()\n");

   /* derive the host URL from the HTTP referrer field */
   cptr = CgiHttpRefererPtr;
   while (*cptr && *(unsigned short*)cptr != '//') cptr++;
   if (*cptr) cptr += 2;
   while (*cptr && *cptr != '/') cptr++;
   HostUrlLength = cptr - CgiHttpRefererPtr;

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Script-Control: X-content-encoding-gzip=0\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<body onload=\"parent.mdsInitialStart();\">\n");
   fflush (stdout);

   for(;;) sys$hiber();
   /* should never ... */
} 

/*****************************************************************************/
/*
Used for both HTTP and WebSocket requests.
Writes the data update JavaScript into the buffer.
This buffered JavaScript is subsequently written to the client(s).
*/

void MonDeSiFormatData (SYSTEM_DATA *sysptr)

{
   static $DESCRIPTOR (DateTimeFaoDsc, "!17%D\0");
   static $DESCRIPTOR (UpTimeFaoDsc, "!10%D\0");

   int  cnt, status,
        HostUrlLength;
   unsigned long  UpBinTime [2];
   char  *cptr, *sptr, *zptr,
         *TimeStampPtr;
   char  CpuActive [32],
         DateTime [32],
         MemSize [32],
         Stupid_njobcnt [16],
         UpTime [32],
         WasdUsing [128];
   $DESCRIPTOR (DateTimeDsc, DateTime);
   $DESCRIPTOR (UpTimeDsc, UpTime);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiFormatData()\n");

   if (!sysptr->PrevInit)
   {
      sysptr->PrevInit = TRUE;

      /* initialise the previous value storage */
      sysptr->PrevBufferedIO = -1;
      sysptr->PrevBusyCpu = -1;
      sysptr->PrevClusterMsgSent = -1;
      sysptr->PrevClusterMsgRcvd = -1;
      sysptr->PrevClusterKbyteMap = -1;
      sysptr->PrevClusterKbyteSent = -1;
      sysptr->PrevClusterKbyteRcvd = -1;
      sysptr->PrevClusterNodes = 0;
      sysptr->PrevDirectIO = -1;
      sysptr->PrevNetIntBlocksRx = -1;
      sysptr->PrevNetIntBlocksTx = -1;
      sysptr->PrevNetIntBytesRx = -1;
      sysptr->PrevNetIntBytesTx = -1;
      sysptr->PrevFcpRead = -1;
      sysptr->PrevFcpWrite = -1;
      sysptr->PrevLockActivity = -1;
      sysptr->PrevLockWait = -1;
      sysptr->PrevMemInUse = -1;
      sysptr->PrevMinute = -1;
      sysptr->PrevMscpRead = -1;
      sysptr->PrevMscpWrite = -1;
      sysptr->PrevPageFaults = -1;
      sysptr->PrevPageFilePercent = -1;
      sysptr->PrevProcessCount = -1;
      sysptr->PrevProcessIntCount = -1;
      sysptr->PrevUserCpu = -1;

      if (sysptr->DiskCount)
      {
         sysptr->PrevMountCount = -1;
         for (cnt = 0; cnt < DISK_DEVICE_MAX; cnt++)
            sysptr->DiskPrevOperCount[cnt] = -1;
      }
   }

   zptr = (sptr = sysptr->OutputBuffer) + sizeof(sysptr->OutputBuffer)-256;

   if (sysptr->PrevMinute != CurrentNumTime[4] ||
       sysptr->ProcessCount != sysptr->PrevProcessCount ||
       sysptr->MountCount != sysptr->PrevMountCount ||
       sysptr->ClusterNodes != sysptr->PrevClusterNodes ||
       sysptr->PrevWebSockConnectedCount != WebSockConnectedCount ||
       NotePad[0])
   {
      sysptr->NodeInfoTimeStamp = TimeStamp;
      sysptr->PrevWebSockConnectedCount = WebSockConnectedCount;

      sys$fao (&DateTimeFaoDsc, 0, &DateTimeDsc, &CurrentBinTime);
      lib$sub_times (&CurrentBinTime, &sysptr->BootBinTime, &UpBinTime);
      sys$fao (&UpTimeFaoDsc, 0, &UpTimeDsc, &UpBinTime);

      if (sysptr->ActiveCpuCount == sysptr->AvailCpuCount)
         sprintf (CpuActive, "%d", sysptr->ActiveCpuCount);
      else
         sprintf (CpuActive, "%d/%d", sysptr->ActiveCpuCount,
                                      sysptr->AvailCpuCount);

      if (sysptr->MemSize >= 1024)
         sprintf (MemSize, "%.1f&nbsp;GB", (float)sysptr->MemSize / 1024.0);
      else
         sprintf (MemSize, "%d&nbsp;MB", sysptr->MemSize);

      sptr += sprintf (sptr,
"if(window.mdsNodeNotes)the=window;else the=parent;\
%s\
the.mdsCpuCount=%d;\
the.mdsNodeNotes(\"<span class=\'node_name\'>%s</span>\
<span class=\'node_time\'>%s</span>\
<span class=\'node_text\'>\
%s %s&nbsp;CPU%s %s %s&nbsp;%s&nbsp <nobr>Up %s</nobr>",
                       /* if script restarts process monitor must be reset */ 
                       (signed int)sysptr->PrevProcessCount == -1 ?
                          "the.mdsProcClick(false);" : "",
                       sysptr->ActiveCpuCount,
                       sysptr->NodeName,
                       DateTime,
                       sysptr->HwName,
                       CpuActive,
                       sysptr->AvailCpuCount > 1 ? "s" : "",
                       MemSize,
                       sysptr->ArchName,
                       sysptr->VmsVersion,
                       UpTime);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (sysptr->ClusterMember)
      {
         sptr += sprintf (sptr,
"&nbsp Cluster nodes:%d votes:%d expect:%d quorum:%d",
                          sysptr->ClusterNodes,
                          sysptr->ClusterVotes,
                          sysptr->ClusterEVotes,
                          sysptr->ClusterQuorum);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
         sysptr->PrevClusterNodes = sysptr->ClusterNodes;
      }

      if ((int)sysptr->ProcessNetCount >= 0)
         sprintf (Stupid_njobcnt, "%d", sysptr->ProcessNetCount);
      else
         strcpy (Stupid_njobcnt, "?");

      sptr += sprintf (sptr,
"&nbsp; Processes:%d interactive:%d batch:%d network:%s",
                       sysptr->ProcessCount,
                       sysptr->ProcessIntCount,
                       sysptr->ProcessBatCount,
                       Stupid_njobcnt);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (sysptr->DiskCount)
      {
         sptr += sprintf (sptr, "&nbsp; Disks:%d/%d",
                          sysptr->MountCount, sysptr->DiskCount);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      }

      if ((int)sysptr->PageFilePercent >= 50)
      {
         sptr += sprintf (sptr, "&nbsp; <span%s>PGFL:%d%%</span>",
                          (int)sysptr->PageFilePercent >= 85 ?
                             " style=\'font-weight:bold;color:red;\'" : "",
                          sysptr->PageFilePercent);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      }

      if (sysptr->NetIntErrorsSoft || sysptr->NetIntErrorsHard)
      {
         sptr += sprintf (sptr, "&nbsp; <span%s>NI:%d/%d</span>",
                          sysptr->NetIntErrorsHard ?
                             " style=\'font-weight:bold;color:red;\'" : "",
                          (int)sysptr->NetIntErrorsSoft,
                          (int)sysptr->NetIntErrorsHard);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      }

      if (WasdHttpd)
      {
         if (WebSockConnectedCount)
            sprintf (WasdUsing, "%s/WebSocket:%d",
                     UsingCGI ? "/CGI" : "", WebSockConnectedCount);
         else
            strcpy (WasdUsing, "/CGI");
      }
      else
         WasdUsing[0] = '\0';

      sptr += sprintf (sptr,
"&nbsp; <span style=\'color:gray\'>(v%s/%s%s%s%s%s)</span>",
                       SOFTWAREVN,
                       ApacheHttpd ? "Apache" : "",
                       OsuHttpd ? "OSU" : "",
                       WasdHttpd ? "WASD" : "",
                       WasdUsing,
                       !(ApacheHttpd || OsuHttpd || WasdHttpd) ?
                          "?httpd?" : "");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (NotePad[0])
      {
         sptr += sprintf (sptr, "&nbsp; <span style=\'color:red;\'>%s</span>",
                          NotePad[0] == '$' ? NotePad+1 : NotePad);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
         if (NotePad[0] != '$') NotePad[0] = '\0';
      }

      sptr += sprintf (sptr, "</span>\");");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (sysptr->DiskCount)
      {
         sptr += sprintf (sptr, "if(the.mdsDiskMount!=%d){the.mdsDiskMount=%d;",
                          sysptr->MountCount, sysptr->MountCount);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
         for (cnt = 0; cnt < sysptr->DiskCount; cnt++)
         {
            sptr += sprintf (sptr,
                             "the.mdsNodeTitle(\"disk_%d\",\"%s %s\");",
                             cnt,
                             sysptr->DiskName[cnt],
                             sysptr->DiskVolName[cnt]);
            if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
         }
         if (sptr < zptr) *sptr++ = '}';
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      }

      sysptr->NodeInfoLength = sptr - sysptr->OutputBuffer;

      /* ensure memory in use and page file is refreshed */
      sysptr->PrevMemInUse = sysptr->PrevPageFilePercent = -1;
   }
   else
   {
      /* leave the node information alone (persists for up to a minute) */
      sptr += sysptr->NodeInfoLength;
   }

   /* only after enough samples to make it representative */
   if (CollectDataSeconds <= 2)
   {
      sysptr->PrevMinute = CurrentNumTime[4];
      sysptr->MonValLength = 0;
      return;
   }

   sptr += sprintf (sptr, "the.mdsTimeStamp=%u;", TimeStamp);
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
   TimeStampPtr = sptr;

   if (sysptr->CpuBusyPercent != sysptr->PrevBusyCpu ||
       sysptr->CpuUserPercent != sysptr->PrevUserCpu)
   {
      sptr += sprintf (sptr,
                       "the.mdsMonVal(\"cpu\",%u,%u,%u);",
                       sysptr->ActiveCpuCount*100,
                       sysptr->CpuBusyPercent,
                       sysptr->CpuBusyPercent - sysptr->CpuUserPercent);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevBusyCpu = sysptr->CpuBusyPercent;
      sysptr->PrevUserCpu = sysptr->CpuUserPercent;
   }

   if (sysptr->PrevMinute != CurrentNumTime[4] ||
       sysptr->MemInUse != sysptr->PrevMemInUse)
   {
      sptr += sprintf (sptr,
#ifndef __VAX
                       "the.mdsMonVal(\"mem\",%llu,%llu);",
                       (__int64)(sysptr->MemSize) << 20,
                       (__int64)(sysptr->MemInUse) << 20);
#else
                       "the.mdsMonVal(\"mem\",%u,%u);",
                       sysptr->MemSize,
                       sysptr->MemInUse);
#endif
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevMemInUse = sysptr->MemInUse;
   }

   if (sysptr->PrevMinute != CurrentNumTime[4] ||
       sysptr->PageFilePercent != sysptr->PrevPageFilePercent)
   {
      sptr += sprintf (sptr,
#ifndef __VAX
                       "the.mdsMonVal(\"pgfl\",%llu,%llu);",
                       (__int64)sysptr->PageFilePage *
                       (__int64)sysptr->PageSize,
                       ((__int64)sysptr->PageFilePage *
                        (__int64)sysptr->PageSize) -
                       ((__int64)sysptr->PageFileFree *
                        (__int64)sysptr->PageSize));
#else
                       "the.mdsMonVal(\"pgfl\",%u,%u);",
                       sysptr->PageFilePage,
                       sysptr->PageFilePage - sysptr->PageFileFree);
#endif
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevPageFilePercent = sysptr->PageFilePercent;
   }

   if (sysptr->PageFaults != sysptr->PrevPageFaults)
   {
      sptr += sprintf (sptr,
                       "the.mdsMonVal(\"flts\",-1,%u,%u);",
                       sysptr->PageFaultsSoft + sysptr->PageFaultsHard,
                       sysptr->PageFaultsHard);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevPageFaults = sysptr->PageFaults;
   }

   if (sysptr->BufferedIO != sysptr->PrevBufferedIO)
   {
      sptr += sprintf (sptr,
                       "the.mdsMonVal(\"bio\",-1,%u);",
                       sysptr->BufferedIO);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevBufferedIO = sysptr->BufferedIO;
   }

   if (sysptr->DirectIO != sysptr->PrevDirectIO)
   {
      sptr += sprintf (sptr,
                       "the.mdsMonVal(\"dio\",-1,%u);",
                       sysptr->DirectIO);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevDirectIO = sysptr->DirectIO;
   }

   if (sysptr->LockActivity != sysptr->PrevLockActivity ||
       sysptr->LockWait != sysptr->PrevLockWait)
   {
      sptr += sprintf (sptr,
                       "the.mdsMonVal(\"lock\",-1,%u,%u);",
                       sysptr->LockActivity,
                       sysptr->LockWait);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevLockActivity = sysptr->LockActivity;
      sysptr->PrevLockWait = sysptr->LockWait;
   }

   if (sysptr->NetIntCount &&
       (sysptr->NetIntBytesRx != sysptr->PrevNetIntBytesRx ||
        sysptr->NetIntBytesTx != sysptr->PrevNetIntBytesTx))
   {
      sptr += sprintf (sptr,
#ifndef __VAX
                       "the.mdsMonVal(\"ni\",-1,%llu,%llu);",
#else
                       "the.mdsMonVal(\"ni\",-1,%u,%u);",
#endif
                       sysptr->NetIntBytesRx + sysptr->NetIntBytesTx,
                       sysptr->NetIntBytesRx);

      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevNetIntBytesRx = sysptr->NetIntBytesRx;
      sysptr->PrevNetIntBytesTx = sysptr->NetIntBytesTx;
   }

   if (sysptr->FcpRead != sysptr->PrevFcpRead ||
       sysptr->FcpWrite != sysptr->PrevFcpWrite)
   {
      sptr += sprintf (sptr,
                       "the.mdsMonVal(\"disk\",-1,%u,%u);",
                       sysptr->FcpRead + sysptr->FcpWrite,
                       sysptr->FcpWrite);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevFcpRead = sysptr->FcpRead;
      sysptr->PrevFcpWrite = sysptr->FcpWrite;
   }

   if (sysptr->PrevMinute != CurrentNumTime[4] ||
       sysptr->ProcessCount != sysptr->PrevProcessCount ||
       sysptr->ProcessIntCount != sysptr->PrevProcessIntCount)
   {
      sptr += sprintf (sptr,
                       "the.mdsMonVal(\"prc\",-1,%u,%u);",
                       sysptr->ProcessCount,
                       sysptr->ProcessIntCount);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevProcessCount = sysptr->ProcessCount;
      sysptr->PrevProcessIntCount = sysptr->ProcessIntCount;
   }

   if (sysptr->ClusterMember &&
       (sysptr->ClusterMsgRcvd != sysptr->PrevClusterMsgRcvd ||
        sysptr->ClusterMsgSent != sysptr->PrevClusterMsgSent))
   {
      sptr += sprintf (sptr,
                       "the.mdsMonVal(\"scsmg\",-1,%u,%u);",
                       sysptr->ClusterMsgRcvd + sysptr->ClusterMsgSent,
                       sysptr->ClusterMsgSent);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevClusterMsgRcvd = sysptr->ClusterMsgRcvd;
      sysptr->PrevClusterMsgSent = sysptr->ClusterMsgSent;
   }

   if (sysptr->ClusterMember &&
       (sysptr->ClusterKbyteMap != sysptr->PrevClusterKbyteMap))
   {
      sptr += sprintf (sptr,
                       "the.mdsMonVal(\"scskb\",-1,%u,%u);",
                       sysptr->ClusterKbyteMap,
                       0);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevClusterKbyteMap = sysptr->ClusterKbyteMap;
   }

   if (sysptr->ClusterMember &&
       (sysptr->MscpRead != sysptr->PrevMscpRead ||
        sysptr->MscpWrite != sysptr->PrevMscpWrite))
   {
      sptr += sprintf (sptr,
                       "the.mdsMonVal(\"mscp\",-1,%u,%u);",
                       sysptr->MscpRead + sysptr->MscpWrite,
                       sysptr->MscpWrite);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sysptr->PrevMscpRead = sysptr->MscpRead;
      sysptr->PrevMscpWrite = sysptr->MscpWrite;
   }

   if (sysptr->DiskCount)
   {
      for (cnt = 0; cnt < sysptr->DiskCount; cnt++)
      {
         if (sysptr->DiskOperCount[cnt] != sysptr->DiskPrevOperCount[cnt])
         {
            sptr += sprintf (sptr,
                             "the.mdsMonVal(\"disk_%d\",-1,%u,%u);",
                             cnt,
                             sysptr->DiskOperCount[cnt],
                             0);
            if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

            sysptr->DiskPrevOperCount[cnt] = sysptr->DiskOperCount[cnt];
         }
      }
      sysptr->PrevMountCount = sysptr->MountCount;
   }

   /* if no data JavaScript follows the timestamp */
   if (sptr == TimeStampPtr) *(sptr = TimeStampPtr) = '\0';

   sysptr->MonValLength = sptr - (sysptr->OutputBuffer+sysptr->NodeInfoLength);
}

/*****************************************************************************/
/*
Indepdendent, timer AST driven update of the system performance data.
This will support multiple clients more efficiently.
The updated data are then read by other indepdendent client threads.
*/

void CollectData ()

{
   static const long cvtf_mode = LIB$K_DELTA_SECONDS_F;
   static unsigned long  PrevBinTime [2];
   static int  SecondsDelta [2] = { 0, -1 };

   int  status;
   unsigned long  DiffBinTime [2];
   struct WsLibStruct  *wsctx, *wsptr;
   CLIENT_DATA  *clptr;
   SYSTEM_DATA  *sysptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectData()\n");

   /* one to sixty seconds, just multiply by interval */
   if (!SecondsDelta[0]) SecondsDelta[0] = -10000000 * RefreshInterval;

   sys$gettim (&CurrentBinTime);
   sys$numtim (&CurrentNumTime, &CurrentBinTime);
   TimeStamp = decc$fix_time (&CurrentBinTime);

   if (ExitTimeStamp && TimeStamp > ExitTimeStamp) exit (SS$_NORMAL); 

   /* calculate a floating point delta time */
   if (!PrevBinTime[0] && !PrevBinTime[1])
   {
      PrevBinTime[0] = CurrentBinTime[0];
      PrevBinTime[1] = CurrentBinTime[1];
   }
   lib$sub_times (&CurrentBinTime, &PrevBinTime, &DiffBinTime);
   PrevBinTime[0] = CurrentBinTime[0];
   PrevBinTime[1] = CurrentBinTime[1];
#ifdef __ia64
   lib$cvts_from_internal_time (&cvtf_mode, &DeltaSeconds, &DiffBinTime);
#else
   lib$cvtf_from_internal_time (&cvtf_mode, &DeltaSeconds, &DiffBinTime);
#endif
   if (Debug) fprintf (stdout, "DeltaSeconds: %f\n", DeltaSeconds);

   sysptr = &SystemData;

   if (sysptr->ProcCount)
   {
      /* determine whether process data should be collected (expensive!) */
      sysptr->ProcMonitor = 0;
      if (UsingCGI)
         sysptr->ProcMonitor = 1;
      else
      {
         /* WebSocket access */
         wsctx = NULL;
         while (wsptr = WsLibNext(&wsctx))
         {
            clptr = WsLibGetUserData (wsptr);
            if (clptr->ProcMonitor) sysptr->ProcMonitor = 1;
         }
      }
   }

   CollectSyiDynamic (sysptr);
   CollectRMI (sysptr);
   CollectNetInt (sysptr);
   CollectDiskSPI (sysptr);
   (*sysptr->ProcSpiFunction)(sysptr);

   MonDeSiFormatData (sysptr);

   if (UsingCGI)
   {
      /* CGI(plus) access */
      if (sysptr->MonValLength ||
          CgiNodeInfoTimeStamp != sysptr->NodeInfoTimeStamp)
      {
         fputs ("<script language=\"JavaScript\">", stdout);

         if (CgiNodeInfoTimeStamp != sysptr->NodeInfoTimeStamp)
         {
            fputs (sysptr->OutputBuffer, stdout);
            CgiNodeInfoTimeStamp = sysptr->NodeInfoTimeStamp;
         }
         else
         if (sysptr->MonValLength)
            fputs (sysptr->OutputBuffer+sysptr->NodeInfoLength, stdout);

         if (sysptr->ProcessBufferLength)
            fputs (sysptr->ProcessBuffer, stdout);

         fputs ("</script>", stdout);
         fflush (stdout);
      }
   }

   /* WebSocket access */
   wsctx = NULL;
   while (wsptr = WsLibNext(&wsctx))
   {
      clptr = WsLibGetUserData (wsptr);

      if (clptr->NodeInfoTimeStamp != sysptr->NodeInfoTimeStamp)
      {
         WsLibWrite (wsptr,
                     sysptr->OutputBuffer,
                     sysptr->NodeInfoLength+sysptr->MonValLength,
                     WSLIB_ASYNCH);
         clptr->NodeInfoTimeStamp = sysptr->NodeInfoTimeStamp;
      }
      else
      if (sysptr->MonValLength)
         WsLibWrite (wsptr,
                     sysptr->OutputBuffer+sysptr->NodeInfoLength,
                     sysptr->MonValLength,
                     WSLIB_ASYNCH);

      if (clptr->ProcMonitor &&
          sysptr->ProcessBufferLength)
         WsLibWrite (wsptr,
                     sysptr->ProcessBuffer,
                     sysptr->ProcessBufferLength,
                     WSLIB_ASYNCH);
   }

   sysptr->PrevMinute = CurrentNumTime[4];

   status = sys$setimr (0, &SecondsDelta, CollectData, NULL, 0);

   if (VMSnok(status)) EXIT_FI_LI (status);

   CollectDataSeconds += RefreshInterval;
}

/*****************************************************************************/
/*
Per-boot system data.
*/ 

int CollectSyiStatic (SYSTEM_DATA *sysptr)

{
   static unsigned long  ActiveCpuCnt,
                         AvailCpuCount,
                         MemSize,
                         PageSize;
   static unsigned long  BootBinTime [2];
   static unsigned short  HwNameLength;
   static unsigned char  ClusterMember;
   static char  ArchName [15+1],
                HwName [1+60],
                NodeName [15+1],
                VmsVersion [8+1];

   /* initialize structure for SYI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList[] =
   {
       { sizeof(AvailCpuCount), SYI$_AVAILCPU_CNT, &AvailCpuCount, 0 },
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(ArchName), SYI$_ARCH_NAME, &ArchName, 0 },
       { sizeof(BootBinTime), SYI$_BOOTTIME, &BootBinTime, 0 },
       { sizeof(ClusterMember), SYI$_CLUSTER_MEMBER, &ClusterMember, 0 },
       { sizeof(HwName), SYI$_HW_NAME, &HwName, &HwNameLength },
       { sizeof(MemSize), SYI$_MEMSIZE, &MemSize, 0 },
       { sizeof(NodeName), SYI$_SCSNODE, &NodeName, 0 },
       { sizeof(PageSize), SYI$_PAGE_SIZE, &PageSize, 0 },
       { sizeof(VmsVersion), SYI$_VERSION, &VmsVersion, 0 },
       {0,0,0,0}
   };

   int  status;
   char  *cptr, *sptr, *zptr;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectSyiStatic()\n");

   status = sys$getsyiw (
            0,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            &ItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getsyi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSnok (status)) return (status);

   zptr = (sptr = sysptr->ArchName) + sizeof(sysptr->ArchName)-1;
   for (cptr = ArchName;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   HwName[HwNameLength] = '\0';
   zptr = (sptr = sysptr->HwName) + sizeof(sysptr->HwName)-1;
   for (cptr = HwName; *cptr && sptr < zptr; *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = sysptr->NodeName) + sizeof(sysptr->NodeName)-1;
   for (cptr = NodeName;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = sysptr->VmsVersion) + sizeof(sysptr->VmsVersion)-1;
   for (cptr = VmsVersion;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   if (!(sysptr->ClusterMember = ClusterMember))
      if (cptr = getenv("MONDESI_CLUSTER_MEMBER"))
         sysptr->ClusterMember = atoi(cptr);

   /* convert to MBytes */
   sysptr->MemSize = MemSize >> 7;
   sysptr->PageSize = PageSize;

   sysptr->BootBinTime[0] = BootBinTime[0];
   sysptr->BootBinTime[1] = BootBinTime[1];

   sysptr->AvailCpuCount = AvailCpuCount;
   sysptr->ActiveCpuCount = ActiveCpuCnt;

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%s| %d %d\n",
               sysptr->ArchName, sysptr->HwName,
               sysptr->NodeName, sysptr->VmsVersion,
               sysptr->AvailCpuCount, ClusterMember);

   return (status);
} 

/*****************************************************************************/
/*
Dynamic system data.
*/ 

int CollectSyiDynamic (SYSTEM_DATA *sysptr)

{
   static unsigned long  ActiveCpuCnt,
                         PageFileFree,
                         PageFilePage;
   static unsigned short  ClusterEVotes,
                          ClusterNodes,
                          ClusterQuorum,
                          ClusterVotes;

   /* initialize structure for SYI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ClusterItemList[] =
   {
       { sizeof(ClusterNodes), SYI$_CLUSTER_NODES, &ClusterNodes, 0 },
       { sizeof(ClusterQuorum), SYI$_CLUSTER_QUORUM, &ClusterQuorum, 0 },
       { sizeof(ClusterVotes), SYI$_CLUSTER_VOTES, &ClusterVotes, 0 },
       { sizeof(ClusterEVotes), SYI$_CLUSTER_EVOTES, &ClusterEVotes, 0 },
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(PageFileFree), SYI$_PAGEFILE_FREE, &PageFileFree, 0 },
       { sizeof(PageFilePage), SYI$_PAGEFILE_PAGE, &PageFilePage, 0 },
       {0,0,0,0}
   }, NonItemList[] =
   {
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(PageFileFree), SYI$_PAGEFILE_FREE, &PageFileFree, 0 },
       { sizeof(PageFilePage), SYI$_PAGEFILE_PAGE, &PageFilePage, 0 },
       {0,0,0,0}
   };

   int  status;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectSyiDynamic()\n");

   /* collect System Information */
   status = sys$getsyiw (
            0,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            sysptr->ClusterMember ?
               (void*)&ClusterItemList :
               (void*)&NonItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getsyi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
      return (status);
   }

   if (sysptr->ClusterMember)
   {
      sysptr->ClusterEVotes = ClusterEVotes;
      sysptr->ClusterNodes = ClusterNodes;
      sysptr->ClusterQuorum = ClusterQuorum;
      sysptr->ClusterVotes = ClusterVotes;
   }

   sysptr->ActiveCpuCount = ActiveCpuCnt;

   sysptr->PageFilePercent = 100 - (PageFileFree * 100 / PageFilePage);
   sysptr->PageFileFree = PageFileFree;
   sysptr->PageFilePage = PageFilePage;

   return (status);
} 

/*****************************************************************************/
/*
Collect System Performance Information.

Uses the system service $GETRMI() available for VMS V7.3 and later.

Cannot use RMI$_CPU* or RMI$_PROC*CNT on VMS V7.3-1 (and V7.3?) because they
don't seem to exist!  Fall back to handcrafting CPU from RMI$_MODES and
sys$gl_*jobcnt (many thanks to Graham Burley for the suggestion).  This
approach maintains two item lists and if the V7.3-2 (and later) fails with a
SS$$_BADPARAM switches to the other list.  Subsequent alternate processing is
performed by testing which list is in use.
*/ 

int CollectRMI (SYSTEM_DATA *sysptr)

{
/* required for VMS V7.3-1 and earlier */
#define INTSTK_MODE  0
#define MPSYNCH_MODE 1
#define KERNEL_MODE  2
#define EXEC_MODE    3
#define SUPER_MODE   4
#define USER_MODE    5
/* 6 is reserved */
#define IDLE_MODE    7

#define MODE_COUNT IDLE_MODE+1
#define CPU_MAX             32

   /* structure for a single CPU's mode ticks */
   static struct SingleCPU {
       unsigned char  CPUid;
       unsigned long  Count [MODE_COUNT];
   };

   /* a structure all CPU's mode ticks for a system with up to 32 CPUs */
   static struct {
       unsigned long  NumberOfCPUs;
       struct SingleCPU  CPU [CPU_MAX];
   } ModeTicks;

   static unsigned long  BufferedIO,
                         ClusterMsgSent,
                         ClusterMsgRcvd,
                         ClusterKbyteMap,
                         ClusterKbyteSent,
                         ClusterKbyteRcvd,
                         DirectIO,
                         FcpRead,
                         FcpWrite,
                         LockConvert,
                         LockDequeue,
                         LockEnqueue,
                         LockWait,
                         MscpRead,
                         MscpWrite,
                         PageFaults,
                         PageReadIO,
                         PageWriteIO,
                         ProcessCount,
                         ProcessBatCount,
                         ProcessIntCount,
                         ProcessNetCount,
                         ScsDataSize;

   static unsigned long  PrevBufferedIO,
                         PrevClusterMsgSent,
                         PrevClusterMsgRcvd,
                         PrevClusterKbyteMap,
                         PrevClusterKbyteSent,
                         PrevClusterKbyteRcvd,
                         PrevDirectIO,
                         PrevFcpRead,
                         PrevFcpWrite,
                         PrevLockActivity,
                         PrevLockWait,
                         PrevMscpRead,
                         PrevMscpWrite,
                         PrevPageFaults,
                         PrevPageReadIO,
                         PrevPageWriteIO,
                         PrevMscpIO;

   static unsigned long  CpuExec [2],
                         CpuIntStk [2],
                         CpuKernel [2],
                         CpuMpSynch [2],
                         CpuSuper [2],
                         CpuUser [2];

   static unsigned long  MscpData [35];

   static unsigned long  PrevCpuTicksBusy [2],
                         PrevCpuTicksUser [2];

   static unsigned long  *ScsDataPtr;

#ifdef __ia64
   static unsigned long  FreeList [2],
                         ModList [2];
#else
   static unsigned long  FreeList,
                         ModList;
#endif

   /* initialize structure for RMI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList1[] =
   {
       /* for 7.3-2(?) and later which has the RMI$_CPU_... items */
       { sizeof(BufferedIO), RMI$_BUFIO, &BufferedIO, 0 },
       { sizeof(CpuExec), RMI$_CPUEXEC, &CpuExec, 0 },
       { sizeof(CpuIntStk), RMI$_CPUINTSTK, &CpuIntStk, 0 },
       { sizeof(CpuKernel), RMI$_CPUKERNEL, &CpuKernel, 0 },
       { sizeof(CpuMpSynch), RMI$_CPUMPSYNCH, &CpuMpSynch, 0 },
       { sizeof(CpuSuper), RMI$_CPUSUPER, &CpuSuper, 0 },
       { sizeof(CpuUser), RMI$_CPUUSER, &CpuUser, 0 },
       { sizeof(LockDequeue), RMI$_DEQ, &LockDequeue, 0 },
       { sizeof(DirectIO), RMI$_DIRIO, &DirectIO, 0 },
       { sizeof(LockConvert), RMI$_ENQCVT, &LockConvert, 0 },
       { sizeof(LockEnqueue), RMI$_ENQNEW, &LockEnqueue, 0 },
       { sizeof(LockWait), RMI$_ENQWAIT, &LockWait, 0 },
       { sizeof(FcpRead), RMI$_FCPREAD, &FcpRead, 0 }, 
       { sizeof(FcpWrite), RMI$_FCPWRITE, &FcpWrite, 0 }, 
       { sizeof(FreeList), RMI$_FRLIST, &FreeList, 0 }, 
       { sizeof(ModList), RMI$_MODLIST, &ModList, 0 },
       { sizeof(MscpData), RMI$_MSCP_EVERYTHING, &MscpData, 0 },
       { sizeof(PageFaults), RMI$_FAULTS, &PageFaults, 0 },
       { sizeof(PageReadIO), RMI$_PREADIO, &PageReadIO, 0 },
       { sizeof(PageWriteIO), RMI$_PWRITIO, &PageWriteIO, 0 },
       { sizeof(ProcessCount), RMI$_PROCS, &ProcessCount, 0 },
       { sizeof(ProcessIntCount), RMI$_PROCINTCNT, &ProcessIntCount, 0 },
       { sizeof(ProcessBatCount), RMI$_PROCBATCNT, &ProcessBatCount, 0 },
       { sizeof(ProcessNetCount), RMI$_PROCNETCNT, &ProcessNetCount, 0 },
       {0,0,0,0}
   }, ItemList2[] =
   {
       /* for 7.3-1(?) and earlier which do not have RMI$_CPU_... items */
       { sizeof(BufferedIO), RMI$_BUFIO, &BufferedIO, 0 },
       { sizeof(ModeTicks), RMI$_MODES, &ModeTicks, 0 },
       { sizeof(LockDequeue), RMI$_DEQ, &LockDequeue, 0 },
       { sizeof(DirectIO), RMI$_DIRIO, &DirectIO, 0 },
       { sizeof(LockConvert), RMI$_ENQCVT, &LockConvert, 0 },
       { sizeof(LockEnqueue), RMI$_ENQNEW, &LockEnqueue, 0 },
       { sizeof(LockWait), RMI$_ENQWAIT, &LockWait, 0 },
       { sizeof(FcpRead), RMI$_FCPREAD, &FcpRead, 0 }, 
       { sizeof(FcpWrite), RMI$_FCPWRITE, &FcpWrite, 0 }, 
       { sizeof(FreeList), RMI$_FRLIST, &FreeList, 0 }, 
       { sizeof(ModList), RMI$_MODLIST, &ModList, 0 },
       { sizeof(PageFaults), RMI$_FAULTS, &PageFaults, 0 },
       { sizeof(PageReadIO), RMI$_PREADIO, &PageReadIO, 0 },
       { sizeof(PageWriteIO), RMI$_PWRITIO, &PageWriteIO, 0 },
       { sizeof(ProcessCount), RMI$_PROCS, &ProcessCount, 0 },
       {0,0,0,0}
   }, ItemList3[] =
   {
       { 0, RMI$_SCS, 0, 0 },
       { sizeof(MscpRead), RMI$_MSCP_READ, &MscpRead, 0 },
       { sizeof(MscpWrite), RMI$_MSCP_WRITE, &MscpWrite, 0 },
       {0,0,0,0}
   }, *ItemListPtr = ItemList1;

   int  cidx, midx, status, tlong;
   unsigned long  datsize;
   unsigned long  CpuTicksBusy [2];
   unsigned long  *datptr;
   float  tfloat;
   IOSBLK  IOsb,
           SynchIOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectRMI()\n");

   /* collect System Performance Information */
   status = sys$getrmi (
            SpiEfn,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            ItemListPtr,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getrmi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   /* fall back to using the pre VMS V7.3-2 items */ 
   if (status == SS$_BADPARAM || IOsb.iosb$w_status == SS$_BADPARAM)
   {
      if (ItemListPtr == ItemList1)
      {
         ItemListPtr = ItemList2;
         return (status);
      }
   }

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
      return (status);
   }

   if (sysptr->ClusterMember)
   {
      /* collect cluster information */
      datsize = ((sysptr->ClusterNodes * 14) + 1) * sizeof(unsigned long);
      if (Debug) fprintf (stdout, "%d %d\n", sysptr->ClusterNodes, datsize);
      if (ScsDataSize < datsize)
      {
         if (ScsDataPtr) free (ScsDataPtr);
         ScsDataPtr = (unsigned long*)calloc (1, ScsDataSize = datsize);
         if (!ScsDataPtr) EXIT_FI_LI (vaxc$errno);
      }
      ItemList3[0].BufferPtr = ScsDataPtr;
      ItemList3[0].BufferLength = ScsDataSize;

      /* does not appear to be supported by $GETRMI */
      status = exe$getspi (
               SpiEfn,  /* efn */
               0,  /* csiaddr */
               0,  /* nodename */
               &ItemList3,  /* item list */
               &IOsb,  /* iosb */
               0,      /* astaddr */
               0 );    /* astprm */

      if (Debug)
         fprintf (stdout, "exe$getspi() %%X%08.08X  IOsb: %%X%08.08X\n",
                  status, IOsb.iosb$w_status);

      if (VMSnok (status))
      {
         sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
         return (status);
      }

      /*** huh? ***/
      ScsDataPtr = ItemList3[0].BufferPtr;

      if (Debug) fprintf (stdout, "%d\n", ScsDataPtr[0]);
      if (ScsDataPtr[0] > sysptr->ClusterNodes)  EXIT_FI_LI (SS$_BUGCHECK);

      for (cidx = 0; cidx < ScsDataPtr[0]; cidx++)
      {
         datptr = &(ScsDataPtr[(cidx*14)+1]);
         if (Debug)
            fprintf (stdout, "%d |%s| %d %d %d %d %d %d %d %d %d %d %d\n",
                     cidx, (char*)(&datptr[0])+1,
                     datptr[2], datptr[3], datptr[4], datptr[5],
                     datptr[6], datptr[7], datptr[8], datptr[9],
                     datptr[10], datptr[11], datptr[12]);
         ClusterMsgSent = datptr[5];
         ClusterMsgRcvd = datptr[6];
         ClusterKbyteSent = datptr[8];
         ClusterKbyteRcvd = datptr[10];
         ClusterKbyteMap = datptr[11];
      }
  }

/***
   status = sys$synch (SpiEfn, &SynchIOsb);

   if (Debug)
      fprintf (stdout, "sys$synch() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, SynchIOsb.iosb$w_status);

   if (VMSnok (status)) return (status);
***/

   /*******/
   /* CPU */
   /*******/

   if (ItemListPtr == ItemList1)
   {
      /* VMS 7.3-2 and later */
      status = lib$addx (&CpuExec, &CpuIntStk, &CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);
      status = lib$addx (&CpuKernel, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);
      status = lib$addx (&CpuMpSynch, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);
      status = lib$addx (&CpuSuper, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);
      status = lib$addx (&CpuUser, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);
   }
   else
   {
     /* VMS V7.3-1 and earlier */
     int  cidx, midx;
     for (cidx = 1; cidx < ModeTicks.NumberOfCPUs; cidx++)
        for (midx = INTSTK_MODE; midx <= IDLE_MODE; midx++)
            ModeTicks.CPU[0].Count[midx] += ModeTicks.CPU[cidx].Count[midx];

     CpuTicksBusy[1] = CpuUser[1] = 0;
     CpuTicksBusy[0] = ModeTicks.CPU[0].Count[INTSTK_MODE];
     /* interrupt item is actually the sum of interrupt and idle! */
     CpuTicksBusy[0] -= ModeTicks.CPU[0].Count[IDLE_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[KERNEL_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[MPSYNCH_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[EXEC_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[SUPER_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[USER_MODE];
     CpuUser[0] = ModeTicks.CPU[0].Count[USER_MODE];

#ifdef  __VAX
     ProcessIntCount = sys$gw_ijobcnt;
     ProcessBatCount = sys$gw_bjobcnt;
     ProcessNetCount = sys$gw_njobcnt;
#else
     ProcessIntCount = sys$gl_ijobcnt;
     ProcessBatCount = sys$gl_bjobcnt;
     ProcessNetCount = sys$gl_njobcnt;
#endif
     /* of course sys$g*_njobcnt doesn't seem to account network jobs! */
     ProcessNetCount = -1;
   }

   if (PrevCpuTicksBusy[0] || PrevCpuTicksBusy[1])
   {
      status = lib$subx (&CpuTicksBusy, &PrevCpuTicksBusy,
                         &sysptr->CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);

      tfloat = ((float)sysptr->CpuTicksBusy[0] +
                ((float)sysptr->CpuTicksBusy[1] * 4294967296.0));
      tfloat /= DeltaSeconds;
      sysptr->CpuBusyPercent = (unsigned int)tfloat;
      if (sysptr->CpuBusyPercent > 100 * sysptr->ActiveCpuCount)
         sysptr->CpuBusyPercent = 100 * sysptr->ActiveCpuCount;
   }
   else
      sysptr->CpuBusyPercent = 0;

   PrevCpuTicksBusy[0] = CpuTicksBusy[0];
   PrevCpuTicksBusy[1] = CpuTicksBusy[1];

   if (PrevCpuTicksUser[0] || PrevCpuTicksUser[1])
   {
      status = lib$subx (&CpuUser, &PrevCpuTicksUser,
                         &sysptr->CpuTicksUser, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);

      tfloat = ((float)sysptr->CpuTicksUser[0] +
                ((float)sysptr->CpuTicksUser[1] * 4294967296.0));
      tfloat /= DeltaSeconds;
      sysptr->CpuUserPercent = (unsigned int)tfloat;
      if (sysptr->CpuUserPercent > 100 * sysptr->ActiveCpuCount)
         sysptr->CpuUserPercent = 100 * sysptr->ActiveCpuCount;
   }
   else
      sysptr->CpuUserPercent = 0;

   PrevCpuTicksUser[0] = CpuUser[0];
   PrevCpuTicksUser[1] = CpuUser[1];

   if (Debug)
      fprintf (stdout, "CPU: %d%% USER: %d%%\n",
               sysptr->CpuBusyPercent, sysptr->CpuUserPercent);

   /***************/
   /* page faults */
   /***************/

   if (PrevPageFaults && PageFaults > PrevPageFaults)
      sysptr->PageFaults = PageFaults - PrevPageFaults;
   else
      sysptr->PageFaults = 0;
   PrevPageFaults = PageFaults;

   if (PrevPageReadIO && PageReadIO > PrevPageReadIO)
      sysptr->PageReadIO = PageReadIO - PrevPageReadIO;
   else
      sysptr->PageReadIO = 0;
   PrevPageReadIO = PageReadIO;

   if (PrevPageWriteIO && PageWriteIO > PrevPageWriteIO)
      sysptr->PageWriteIO = PageWriteIO - PrevPageWriteIO;
   else
      sysptr->PageWriteIO = 0;
   PrevPageWriteIO = PageWriteIO;

   sysptr->PageFaultsHard = sysptr->PageReadIO + sysptr->PageWriteIO;
   tfloat = (float)sysptr->PageFaultsHard / DeltaSeconds;
   sysptr->PageFaultsHard = (unsigned int)tfloat;

   sysptr->PageFaultsSoft = sysptr->PageFaults - sysptr->PageFaultsHard;
   tfloat = (float)sysptr->PageFaultsSoft / DeltaSeconds;
   sysptr->PageFaultsSoft = (unsigned int)tfloat;

   /********/
   /* MSCP */
   /********/

   if (sysptr->ClusterMember)
   {
      if (ItemListPtr == ItemList1)
      {
         MscpRead = MscpData[23];
         MscpWrite = MscpData[24];
      }

      if (PrevMscpRead && MscpRead > PrevMscpRead)
         sysptr->MscpRead = MscpRead - PrevMscpRead;
      else
         sysptr->MscpRead = 0;
      PrevMscpRead = MscpRead;
      tfloat = (float)sysptr->MscpRead / DeltaSeconds;
      sysptr->MscpRead = (unsigned int)tfloat;

      if (PrevMscpWrite > MscpWrite > PrevMscpWrite)
         sysptr->MscpWrite = MscpWrite - PrevMscpWrite;
      else
         sysptr->MscpWrite = 0;
      PrevMscpWrite = MscpWrite;
      tfloat = (float)sysptr->MscpWrite / DeltaSeconds;
      sysptr->MscpWrite = (unsigned int)tfloat;
   }

   /********/
   /* lock */
   /********/

   if (PrevLockActivity &&
       LockEnqueue + LockDequeue + LockWait > PrevLockActivity)
      sysptr->LockActivity = LockEnqueue + LockDequeue + LockWait -
                             PrevLockActivity;
   else
      sysptr->LockActivity = 0;
   PrevLockActivity = LockEnqueue + LockDequeue + LockWait;
   tfloat = (float)sysptr->LockActivity / DeltaSeconds;
   sysptr->LockActivity = (unsigned int)tfloat;
   tfloat = (float)sysptr->LockWait / DeltaSeconds;
   sysptr->LockWait = (unsigned int)tfloat;

   if (PrevLockWait && LockWait > PrevLockWait)
      sysptr->LockWait = LockWait - PrevLockWait;
   else
      sysptr->LockWait = 0;
   PrevLockWait = LockWait;

   /********/
   /* disk */
   /********/

   if (PrevFcpRead && FcpRead > PrevFcpRead)
      sysptr->FcpRead = FcpRead - PrevFcpRead;
   else
      sysptr->FcpRead = 0;
   PrevFcpRead = FcpRead;
   tfloat = (float)sysptr->FcpRead / DeltaSeconds;
   sysptr->FcpRead = (unsigned int)tfloat;

   if (PrevFcpWrite && FcpWrite > PrevFcpWrite)
      sysptr->FcpWrite = FcpWrite - PrevFcpWrite;
   else
      sysptr->FcpWrite = 0;
   PrevFcpWrite = FcpWrite;
   tfloat = (float)sysptr->FcpWrite / DeltaSeconds;
   sysptr->FcpWrite = (unsigned int)tfloat;

   /*********/
   /* other */
   /*********/

   if (PrevDirectIO && DirectIO > PrevDirectIO)
      sysptr->DirectIO = DirectIO - PrevDirectIO;
   else
      sysptr->DirectIO = 0;
   PrevDirectIO = DirectIO;
   tfloat = (float)sysptr->DirectIO / DeltaSeconds;
   sysptr->DirectIO = (unsigned int)tfloat;

   if (PrevBufferedIO && BufferedIO > PrevBufferedIO)
      sysptr->BufferedIO = BufferedIO - PrevBufferedIO;
   else
      sysptr->BufferedIO = 0;
   PrevBufferedIO = BufferedIO;
   tfloat = (float)sysptr->BufferedIO / DeltaSeconds;
   sysptr->BufferedIO = (unsigned int)tfloat;

#ifdef __ia64
   sysptr->MemInUse = sysptr->MemSize - (FreeList[0] >> 7) - (ModList[0] >> 7);
#else
   sysptr->MemInUse = sysptr->MemSize - (FreeList >> 7) - (ModList >> 7);
#endif
   if (sysptr->MemInUse > sysptr->MemInUseMax)
      sysptr->MemInUseMax = sysptr->MemInUse;

   sysptr->ProcessCount = ProcessCount;
   sysptr->ProcessBatCount = ProcessBatCount;
   sysptr->ProcessIntCount = ProcessIntCount;
   sysptr->ProcessNetCount = ProcessNetCount;

   if (sysptr->ClusterMember)
   {
      /*******/
      /* SCS */
      /*******/

      if (PrevClusterMsgSent && ClusterMsgSent > PrevClusterMsgSent)
         sysptr->ClusterMsgSent = ClusterMsgSent - PrevClusterMsgSent;
      else
         sysptr->ClusterMsgSent = 0;
      PrevClusterMsgSent = ClusterMsgSent;
      tfloat = (float)sysptr->ClusterMsgSent / DeltaSeconds;
      sysptr->ClusterMsgSent = (unsigned int)tfloat;

      if (PrevClusterMsgRcvd && ClusterMsgRcvd > PrevClusterMsgRcvd)
         sysptr->ClusterMsgRcvd = ClusterMsgRcvd - PrevClusterMsgRcvd;
      else
         sysptr->ClusterMsgRcvd = 0;
      PrevClusterMsgRcvd = ClusterMsgRcvd;
      tfloat = (float)sysptr->ClusterMsgRcvd / DeltaSeconds;
      sysptr->ClusterMsgRcvd = (unsigned int)tfloat;

      if (PrevClusterKbyteMap && ClusterKbyteMap > PrevClusterKbyteMap)
         sysptr->ClusterKbyteMap = ClusterKbyteMap - PrevClusterKbyteMap;
      else
         sysptr->ClusterKbyteMap = 0;
      PrevClusterKbyteMap = ClusterKbyteMap;
      tfloat = (float)sysptr->ClusterKbyteMap / DeltaSeconds;
      sysptr->ClusterKbyteMap = (unsigned int)tfloat;

      if (PrevClusterKbyteSent && ClusterKbyteSent > PrevClusterKbyteSent)
         sysptr->ClusterKbyteSent = ClusterKbyteSent - PrevClusterKbyteSent;
      else
         sysptr->ClusterKbyteSent = 0;
      PrevClusterKbyteSent = ClusterKbyteSent;
      tfloat = (float)sysptr->ClusterKbyteSent / DeltaSeconds;
      sysptr->ClusterKbyteSent = (unsigned int)tfloat;

      if (PrevClusterKbyteRcvd && ClusterKbyteRcvd > PrevClusterKbyteRcvd)
         sysptr->ClusterKbyteRcvd = ClusterKbyteRcvd - PrevClusterKbyteRcvd;
      else
         sysptr->ClusterKbyteRcvd = 0;
      PrevClusterKbyteRcvd = ClusterKbyteRcvd;
      tfloat = (float)sysptr->ClusterKbyteRcvd / DeltaSeconds;
      sysptr->ClusterKbyteRcvd = (unsigned int)tfloat;
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
This function uses undocumented sense-mode functionality to load counters from
the network interface (commonly ethernet device).  Multiple devices are
supported using a multi-valued logical name MONDESI_NI and the frame-rate
and byte-rate and soft/hard error-rate statistics are accumlated to represent
total system data.

On Alpha and Itanium uses native 64 bit (quadword) accumulators.

Essential concepts have come from "EMON: moniteur ethernet (V2.1-2) Gerard
Guillaume" and found in DECUSLIB freeware, and from some peeks at the T4 source
code.  The implementation is mine.
*/

int CollectNetInt (SYSTEM_DATA *sysptr)

{
   static int  NetIntCount;
   static unsigned short  NetIntChan [NI_DEVICE_MAX];
   static unsigned long  DcSCOM = 32,  /* DC$_SCOM */
                         DtNI = 13;    /* DT$_NI */

#ifndef __VAX
   static unsigned __int64  PrevBlocksRx,
                            PrevBlocksTx,
                            PrevBytesRx,
                            PrevBytesTx,
                            PrevErrorsHard,
                            PrevErrorsSoft;
#else
   static unsigned long  PrevBlocksRx,
                         PrevBlocksTx,
                         PrevBytesRx,
                         PrevBytesTx,
                         PrevErrorsHard,
                         PrevErrorsSoft;
#endif

   static long  DeviceCount = -1,
                DviUnit,
                LnmIndex = -1,
                LnmAttributes;
   static unsigned short  LogValueLength;
   static char  LogValue [32];
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { sizeof(LogValue)-1, LNM$_STRING, &LogValue, &LogValueLength },
      { 0,0,0,0 }
   },
   DevScanItems [] =
   {
      { sizeof(DcSCOM), DVS$_DEVCLASS, &DcSCOM, 0 },
      { 0,0,0,0 }
   },
   GetDviItems [] =
   {
      { sizeof(DviUnit), DVI$_UNIT, &DviUnit, 0 },
      { 0,0,0,0 }
   };
   static $DESCRIPTOR (NetIntDsc, "");
   static $DESCRIPTOR (LogNameDsc, "MONDESI_NI");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   int  ecnt, status;
   unsigned short  pword,
                   retlen;
   unsigned long  LnmCount,
                  NetIntErrorsHard,
                  NetIntErrorsSoft;
   unsigned long  DevScanContext [2];
   float  tfloat;
   char  *vptr;
   char  DevScanBuffer [64],
         CountBuffer [512];
   struct STRUCT_IOSBLK  IOsb;
   $DESCRIPTOR (CountDsc, CountBuffer);
   $DESCRIPTOR (DevScanDsc, DevScanBuffer);

#ifndef __VAX
#define CASTIT (unsigned __int64)
   unsigned __int64  value,
                     NetIntBlocksRx,
                     NetIntBlocksTx,
                     NetIntBytesRx,
                     NetIntBytesTx;
#else
#define CASTIT (unsigned long)
   unsigned long  value,
                  NetIntBlocksRx,
                  NetIntBlocksTx,
                  NetIntBytesRx,
                  NetIntBytesTx;
#endif

   /*********/
   /* begin */
   /*********/

   if (LnmIndex < 0)
   {
      /* first call, check for logical name defined network devices */
      for (LnmIndex = 0; LnmIndex < NI_DEVICE_MAX; LnmIndex++)
      {
         status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
         if (VMSnok (status) || !(LnmAttributes & LNM$M_EXISTS)) break;
         LogValue[LogValueLength] = '\0';

         NetIntDsc.dsc$a_pointer = LogValue;
         NetIntDsc.dsc$w_length = LogValueLength;

         status = sys$assign (&NetIntDsc, &NetIntChan[LnmIndex], 0, 0);
         if (VMSnok (status))
         {
            sprintf (NotePad, "$[%d] %s %%X%08.08X",
                     __LINE__, LogValue, status);
            NetIntCount = 0;
            break;
         }
         if (NetIntDev[0]) strcat (NetIntDev, ",");
         strcat (NetIntDev, LogValue);
         NetIntCount++;
      }
   }

   if (DeviceCount < 0)
   {
      /* first call */
      DeviceCount = 0;

      /* if logical name defined and found devices */
      if (NetIntCount) return (NetIntCount);

      /* scan system for network devices */
      DevScanContext[0] = DevScanContext[1] = 0;
      while (NetIntCount < NI_DEVICE_MAX)
      {
         status = sys$device_scan (&DevScanDsc, &retlen, 0,
                                   &DevScanItems, &DevScanContext);
         if (VMSnok (status))
         {
            if (status == SS$_NOMOREDEV || status == SS$_NOSUCHDEV) break;
            sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
            NetIntCount = 0;
            break;
         }

         status = sys$getdviw (0, 0, &DevScanDsc, &GetDviItems, &IOsb, 
                               0, 0, 0, 0);
         if (VMSok (status)) status = IOsb.iosb$w_status;
         if (VMSnok (status))
         {
            sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
            NetIntCount = 0;
            break;
         }
         /* if not a template device then continue */
         if (DviUnit != 0) continue;

         status = sys$assign (&DevScanDsc, &NetIntChan[DeviceCount++], 0, 0);
         if (VMSnok (status))
         {
            if (status == SS$_NOPRIV) continue;
            sprintf (NotePad, "$[%d] %s %%X%08.08X",
                     __LINE__, LogValue, status);
            NetIntCount = 0;
            break;
         }

         if (NetIntDev[0]) strcat (NetIntDev, ",");
         DevScanBuffer[retlen] = '\0';
         strcat (NetIntDev, DevScanBuffer+1);
         NetIntCount++;
      }

      /* first call, just return having assigned channels (perhaps) */
      return (NetIntCount);
   }

   /****************/
   /* collect data */
   /****************/

   /* initialise structure and local storage */
   sysptr->NetIntBlocksRx = sysptr->NetIntBlocksTx =
      sysptr->NetIntBytesRx = sysptr->NetIntBytesTx =
      NetIntBlocksRx = NetIntBlocksTx =
      NetIntBytesRx = NetIntBytesTx =
      NetIntErrorsHard = NetIntErrorsSoft = 0;

   /* if no devices then just return from here */
   if (!NetIntCount) return (NetIntCount);

   /* accumulate counts from each device */
   for (ecnt = 0; ecnt < NetIntCount; ecnt++)
   {
      memset (CountBuffer, 0, sizeof(CountBuffer));

      status = sys$qiow (0, NetIntChan[ecnt],
#ifndef __VAX
#define IO$M_RD_64COUNT 0x4000
                         IO$_SENSEMODE | IO$M_RD_COUNT |
                         IO$M_RD_64COUNT | IO$M_CTRL,
#else
                         IO$_SENSEMODE | IO$M_RD_COUNT | IO$M_CTRL,
#endif
                         &IOsb, 0, 0,
                         0, &CountDsc, 0, 0, 0, 0);

      if (VMSok (status)) status = IOsb.iosb$w_status;
      if (VMSnok (status))
      {
         sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
         return (NetIntCount);
      }

      vptr = CountBuffer;
      while (vptr < CountBuffer + IOsb.iosb$w_bcnt)
      {
         pword = *(unsigned short*)vptr;
         vptr += 2;

         /* end of list */
         if (!pword) break;

         /* MSB should be set! */
         if (!(pword & 0x8000))
         {
            sprintf (NotePad, "$[%d] %04.04X", __LINE__, pword);
            break;
         }

         value = 0;
#ifndef __VAX
         if ((pword & 0x0fff) < 200)
         {
            /* quadword value (item number less than 200) */
            value = *(unsigned __int64*)vptr;
            vptr += 8;
         }
         else
#endif
         if ((pword & 0x6000) == 0x6000)
         {
            /* longword value */
            value = *(unsigned long*)vptr;
            vptr += 4;
         } 
         else
         if ((pword & 0x6000) == 0x4000)
         {
            /* word value */
            value = *(unsigned short*)vptr;
            vptr += 2;
         } 
         else
         if ((pword & 0x6000) == 0x2000)
         {
            /* byte value */
            value = *(unsigned char*)vptr;
            vptr++;
         } 

         if (pword & 0x1000)
         {
            /* hw map (?) */
            value = *(unsigned long*)vptr;
            vptr += 2;
            continue;
         } 

         switch (pword & 0x0fff)
         {
#ifndef __VAX
            /* these look suspiciously like "LANCP SHOW DEVICE /COUNT"! */

            case 1 :
                 /* seconds since last zeroed */
                 break;

            case 2 :
            case 6 :
                 NetIntBytesRx += value;
                 break;

            case 3 :
            case 7 :
                 NetIntBytesTx += value;
                 break;

            case 4 :
            case 8 :
                 NetIntBlocksRx += value;
                 break;

            case 5 :
            case 9 :
                 NetIntBlocksTx += value;
                 break;

            case 12 :
            case 13 :
            case 14 :
            case 15 :
            case 16 :
            case 17 :
            case 18 :
            case 19 :
            case 30 :
            case 21 :
            case 22 :
            case 23 :
            case 24 :
            case 27 :
            case 28 :
            case 29 :
                 NetIntErrorsSoft += value;
                 break;

            case 25 :
            case 26 :
            case 33 :
            case 34 :
                 NetIntErrorsHard += value;
                 break;
#endif

            /* more "traditional" counters */

            case 1000 /* NMA$C_CTLIN_BRC */ :
            case 1002 /* NMA$C_CTLIN_MBY */ :
                 NetIntBytesRx += value;
                 break;

            case 1001 /* NMA$C_CTLIN_BSN */ :
                 NetIntBytesTx += value;
                 break;

            case 1010 /* NMA$C_CTLIN_DBR */ :
            case 1012 /* NMA$C_CTLIN_MBL */ :
                 NetIntBlocksRx += value;
                 break;

            case 1011 /* NMA$C_CTLIN_DBS */ :
                 NetIntBlocksTx += value;
                 break;

            case 1013 /* NMA$C_CTLIN_BID */ :
            case 1014 /* NMA$C_CTLIN_BS1 */ :
            case 1015 /* NMA$C_CTLIN_BSM */ :
            case 1040 /* NMA$C_CTLIN_RBE */ :
            case 1041 /* NMA$C_CTLIN_LBE */ :
            case 1064 /* NMA$C_CTLIN_OVR */ :
            case 1066 /* NMA$C_CTLIN_UBU */ :

                 NetIntErrorsSoft += value;
                 break;

            case 1060 /* NMA$C_CTLIN_SFL */ :
            case 1061 /* NMA$C_CTLIN_CDC */ :
            case 1063 /* NMA$C_CTLIN_UFD */ :
            case 1062 /* NMA$C_CTLIN_RFL */ :
            case 1065 /* NMA$C_CTLIN_SBU */ :

                 NetIntErrorsHard += value;
                 break;

            default :
         }
      }
   }

   if (PrevBytesRx && NetIntBytesRx > PrevBytesRx)
      sysptr->NetIntBytesRx = NetIntBytesRx - PrevBytesRx;
   PrevBytesRx = NetIntBytesRx;

   if (PrevBytesTx && NetIntBytesTx > PrevBytesTx)
      sysptr->NetIntBytesTx = NetIntBytesTx - PrevBytesTx;
   PrevBytesTx = NetIntBytesTx;

   tfloat = (float)sysptr->NetIntBytesRx / DeltaSeconds;
   sysptr->NetIntBytesRx = CASTIT tfloat;

   tfloat = (float)sysptr->NetIntBytesTx / DeltaSeconds;
   sysptr->NetIntBytesTx = CASTIT tfloat;

   if (PrevBlocksRx && NetIntBlocksRx > PrevBlocksRx)
      sysptr->NetIntBlocksRx = NetIntBlocksRx - PrevBlocksRx;
   PrevBlocksRx = NetIntBlocksRx;

   if (PrevBlocksTx && NetIntBlocksTx > PrevBlocksTx)
      sysptr->NetIntBlocksTx = NetIntBlocksTx - PrevBlocksTx;
   PrevBlocksTx = NetIntBlocksTx;

   tfloat = (float)sysptr->NetIntBlocksRx / DeltaSeconds;
   sysptr->NetIntBlocksRx = CASTIT tfloat;

   tfloat = (float)sysptr->NetIntBlocksTx / DeltaSeconds;
   sysptr->NetIntBlocksTx = CASTIT tfloat;

   if (PrevErrorsSoft && NetIntErrorsSoft > PrevErrorsSoft)
      sysptr->NetIntErrorsSoft += NetIntErrorsSoft - PrevErrorsSoft;
   PrevErrorsSoft = NetIntErrorsSoft;

   if (PrevErrorsHard && NetIntErrorsHard > PrevErrorsHard)
      sysptr->NetIntErrorsHard += NetIntErrorsHard - PrevErrorsHard;
   PrevErrorsHard = NetIntErrorsHard;

   return (NetIntCount);
}

/*****************************************************************************/
/*
Collect disk System Performance Information.
Uses the system service EXE$GETSPI() as the equivalent SYS$GETRMI() doesn't
seem to implement the functionality.
*/ 

int CollectDiskSPI (SYSTEM_DATA *sysptr)

{
#define GETSPI$_DISKS 4203

typedef struct _getspi$disk_offsets {
    short int getspi$w_disk_alloclass;
    int getspi$l_disk_devname;
    short int getspi$w_disk_unitnum;
    char getspi$b_disk_flags;
    int getspi$q_disk_nodename [2];
    int getspi$q_disk_volnamel [2];
    int getspi$l_disk_volnameh;
    int getspi$l_disk_optcnt;
    int getspi$l_disk_qcount;
    } GETSPI$DISK_OFFSETS;

#define DISK_DATA_SIZE sizeof(unsigned long) + \
                       (sizeof(GETSPI$DISK_OFFSETS)*DISK_OFFSETS_MAX)

   static char  DiskData [DISK_DATA_SIZE];

   /* initialize structure for SPI items */
   static struct
   {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList1[] =
   {
       { sizeof(DiskData), GETSPI$_DISKS, &DiskData, 0 },
       {0,0,0,0}
   };

   static int  DiskCount = 0;
   static long  LnmIndex = -1,
                LnmAttributes;
   static unsigned short  LogValueLength;
   static $DESCRIPTOR (LogNameDsc, "MONDESI_DISK");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { DISK_NAME_SIZE, LNM$_STRING, 0, &LogValueLength },
      { 0,0,0,0 }
   };

   int  dcnt, ncnt, status;
   float  tfloat;
   char  *cptr, *sptr;
   char  DiskName [128];
   struct STRUCT_IOSBLK  IOsb;
   GETSPI$DISK_OFFSETS  *doptr;
   SYSTEM_DATA  DummyData;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectDiskSPI()\n");

   if (LnmIndex < 0)
   {
      for (LnmIndex = 0; LnmIndex < DISK_DEVICE_MAX; LnmIndex++)
      {
         LnmItems[2].BufferPtr = sysptr->DiskName[LnmIndex];
         status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
         if (VMSnok (status) || !(LnmAttributes & LNM$M_EXISTS)) break;
         sysptr->DiskName[LnmIndex][LogValueLength] = '\0';
      }
      if ((sysptr->DiskCount = LnmIndex) == DISK_DEVICE_MAX)
         sysptr->DiskCount--;

      return (sysptr->DiskCount);
   }

   /* if no disks being monitored */
   if (!sysptr->DiskCount) return (0);

   /* collect System Performance Information */
   status = exe$getspi (SpiEfn, 0, 0, &ItemList1, &IOsb, 0, 0);

   if (Debug)
      fprintf (stdout, "exe$getspi() %%X%08.08X  IOsb: %%X%08.08X disk: %d\n",
               status, IOsb.iosb$w_status, *(unsigned long*)&DiskData);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) return (0);

   DiskCount = *(unsigned long*)&DiskData;
   if (DiskCount == DISK_OFFSETS_MAX) EXIT_FI_LI (SS$_BUGCHECK);

   doptr = (GETSPI$DISK_OFFSETS*)(DiskData + sizeof(unsigned long));

   /* if the volume name is not updated the data is not fresh */
   sysptr->MountCount = 0;
   for (ncnt = 0; ncnt < sysptr->DiskCount; ncnt++)
      sysptr->DiskVolName[ncnt][0] = '\0';

   /* transfer from SPI into system data structure */
   for (dcnt = 0; dcnt < DiskCount; dcnt++, doptr++)
   {
      if (doptr->getspi$w_disk_alloclass)
         sprintf (DiskName, "$%d$%*.*s%d:",
                  doptr->getspi$w_disk_alloclass,
                  (int)(((char*)(&doptr->getspi$l_disk_devname))[0]),
                  (int)(((char*)(&doptr->getspi$l_disk_devname))[0]),
                  (char*)&doptr->getspi$l_disk_devname+1,
                  doptr->getspi$w_disk_unitnum);
      else
         sprintf (DiskName, "%*.*s$%*.*s%d:",
                  (int)(((char*)(&doptr->getspi$q_disk_nodename))[0]),
                  (int)(((char*)(&doptr->getspi$q_disk_nodename))[0]),
                  (char*)doptr->getspi$q_disk_nodename+1,
                  (int)(((char*)(&doptr->getspi$l_disk_devname))[0]),
                  (int)(((char*)(&doptr->getspi$l_disk_devname))[0]),
                  (char*)&doptr->getspi$l_disk_devname+1,
                  doptr->getspi$w_disk_unitnum);

      /* locate the particular device in the system data */
      for (ncnt = 0; ncnt < sysptr->DiskCount; ncnt++)
      {
#ifndef __VAX
         /* a fractionally more efficient test */
         if (*(__int64*)sysptr->DiskName[ncnt] != *(__int64*)DiskName) continue;
#endif
         if (!strcmp (sysptr->DiskName[ncnt], DiskName)) break;
      }

      /* if it's not a required disk device */
      if (ncnt >= sysptr->DiskCount) continue;

      /* volume name terminating on first space */
      sptr = sysptr->DiskVolName[ncnt];
      for (cptr = (char*)doptr->getspi$q_disk_volnamel;
           cptr < (char*)(doptr->getspi$q_disk_volnamel)+12 && *cptr != ' ';
           *sptr++ = *cptr++);
      *sptr = '\0';
      if (sysptr->DiskVolName[ncnt][0]) sysptr->MountCount++;

      sysptr->DiskOptCnt[ncnt] = doptr->getspi$l_disk_optcnt;
      if (sysptr->DiskPrevOptCnt[ncnt] &&
          sysptr->DiskOptCnt[ncnt] > sysptr->DiskPrevOptCnt[ncnt])
      {
         tfloat = (float)(sysptr->DiskOptCnt[ncnt] -
                          sysptr->DiskPrevOptCnt[ncnt]) / DeltaSeconds;
         sysptr->DiskOperCount[ncnt] = (unsigned int)tfloat;
      }
      else
         sysptr->DiskOperCount[ncnt] = 0;
      sysptr->DiskPrevOptCnt[ncnt] = sysptr->DiskOptCnt[ncnt];
   }

   return (sysptr->DiskCount);
}

/*****************************************************************************/
/*
Collect process System Performance Information for younger VMSes.
Uses the system service EXE$GETSPI() as the equivalent SYS$GETRMI() doesn't
seem to implement the functionality.

This function collects process SPI and sorts the top five consumers each of CPU
time, super+exec+kernel time, pages faults, buffered and direct I/O.  These
five are then compiled into an HTML table for display purposes.

This function uses a data structure derived in part from VMS sources and
contains items and alignment (seemingly) not documented elsewhere.  Not being
completely obvious when the new structure started to be used (VMS V8.3?) this
function switches between this and the older (documented and header-file
described) data structure by sanity-checking the data returned.  If any of the
item values appear out-of-range then it's likely the offsets are wrong for this
(older) system.  Switch to using an equivalent function that employs that older
structure.
*/ 

int CollectProcSPI1 (SYSTEM_DATA *sysptr)

{
#define GETSPI$_PROC 4123

/* LOTS of trial-and-error in this structure! */

typedef struct _getspi$proc_class {
    int getspi$l_proc_ipid;
    int getspi$l_proc_uic;
    short int getspi$w_proc_state;
    char getspi$b_proc_pri;
    /* OCTAWORD basealign octaword? */
    char filler_to_octaword_alignment [5];
    char getspi$o_proc_lname [16];
    int getspi$l_proc_gpgcnt;
    int getspi$l_proc_ppgcnt;
    int getspi$l_proc_sts;
    int getspi$l_proc_diocnt;
    int getspi$l_proc_pageflts;
    int getspi$l_proc_cputim;
    int getspi$l_proc_biocnt;
    int getspi$l_proc_epid;
    int getspi$l_proc_efwm;
    int getspi$l_proc_rbstran;
    /* from V8.3 IA64 SPIDEF.SDL */
    int getspi$l_proc_kernel_counter;
    int getspi$l_proc_executive_counter;
    int getspi$l_proc_supervisor_counter;
    int getspi$l_proc_user_counter;
    int getspi$l_proc_reserved1;
    /* use the second reserve to accumulate non-user ticks */
    /** int getspi$l_proc_reserved2; **/
    int getspi$l_proc_SEK;
    } GETSPI$PROC_CLASS;

#define PROC_DATA_SIZE sizeof(unsigned long) + \
                       (sizeof(GETSPI$PROC_CLASS)*PROC_CLASS_MAX)

   static long  BufferNumber,
                LnmAttributes;
   static int  DiskCount;
   static unsigned short  PrevMinute = -1;
   static unsigned short  LogValueLength;
   static char  LogValue [32],
                ProcData1 [PROC_DATA_SIZE],
                ProcData2 [PROC_DATA_SIZE];
   static $DESCRIPTOR (LogNameDsc, "MONDESI_PROC");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   /* initialize structure for SPI items */
   static struct
   {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList1[] =
   {
       { sizeof(ProcData1), GETSPI$_PROC, 0, 0 },
       {0,0,0,0}
   };

   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { sizeof(LogValue), LNM$_STRING, LogValue, &LogValueLength },
      { 0,0,0,0 }
   };

   int  cnt, spicnt, sprcnt, status, topcnt,
        PrevProcCount,
        ProcCount,
        RowCount;
   float  tfloat;
   char  *sptr, *zptr;
   unsigned char  ch;
   struct STRUCT_IOSBLK  IOsb;
   GETSPI$PROC_CLASS  *prvptr, *spiptr, *sprptr;
   GETSPI$PROC_CLASS  *TopCpuTim [PROC_MAX],
                      *TopBioCnt [PROC_MAX],
                      *TopDioCnt [PROC_MAX],
                      *TopPageFlts [PROC_MAX],
                      *TopSekTim [PROC_MAX];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectProcSPI1()\n");

   if (!sysptr)
   {
      /* initialise */
      ProcCount = 0;
      status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
      if (VMSok (status) && (LnmAttributes & LNM$M_EXISTS))
      {
         LogValue[LogValueLength] = '\0';
         if (isdigit(LogValue[0]))
            ProcCount = atoi(LogValue);
         else
            ProcCount = PROC_MAX;
      }
      return (ProcCount);
   }

   sysptr->ProcessBuffer[sysptr->ProcessBufferLength = 0] = '\0';

   /* if process not being monitored */
   if (!(sysptr->ProcCount && sysptr->ProcMonitor)) return (0);

   /***************/
   /* collect SPI */
   /***************/

   /* swap back and forth between the two buffers */
   if (BufferNumber++ % 2)
   {
      spiptr = (GETSPI$PROC_CLASS*)ProcData2;
      prvptr = (GETSPI$PROC_CLASS*)ProcData1;
   }
   else
   {
      spiptr = (GETSPI$PROC_CLASS*)ProcData1;
      prvptr = (GETSPI$PROC_CLASS*)ProcData2;
   }
   ItemList1[0].BufferPtr = spiptr;

   /* collect System Performance Information */
   status = exe$getspi (SpiEfn, 0, 0, &ItemList1, &IOsb, 0, 0);

   if (Debug)
      fprintf (stdout, "exe$getspi() %%X%08.08X  IOsb: %%X%08.08X proc: %d\n",
               status, IOsb.iosb$w_status, *(unsigned long*)spiptr);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) return (0);

   if (sysptr->ProcSanityCheck == 0)
   {
      /* sanity check the data structure */
      sysptr->ProcSanityCheck = 1;
      ProcCount = *(unsigned long*)spiptr;
      spiptr = (GETSPI$PROC_CLASS*)((char*)spiptr + sizeof(unsigned long));
      for (spicnt = 0; spicnt < ProcCount; spicnt++, spiptr++)
      {
          /* if any seem out-of-range the offsets are probably wrong */
          if (spiptr->getspi$l_proc_epid == 0 ||
              spiptr->getspi$w_proc_state < 0 ||
              spiptr->getspi$w_proc_state > 255 ||
              *spiptr->getspi$o_proc_lname < 0 ||
              *spiptr->getspi$o_proc_lname > 15)
          {
             /* try using the older version of the structure */
             sysptr->ProcSpiFunction = &CollectProcSPI2;
             return (CollectProcSPI2 (sysptr));
          }
       }
   }

   /**********************/
   /* get process counts */
   /**********************/

   /* number of processes read by SPI */
   ProcCount = *(unsigned long*)spiptr;
   if (ProcCount == PROC_CLASS_MAX) EXIT_FI_LI (SS$_BUGCHECK);
   /* now step over the process count */
   spiptr = (GETSPI$PROC_CLASS*)((char*)spiptr + sizeof(unsigned long));

   /* number of process read into the previous SPI */
   PrevProcCount = *(unsigned long*)prvptr;
   /* now step over the process count */
   prvptr = (GETSPI$PROC_CLASS*)((char*)prvptr + sizeof(unsigned long));
   /* if there is no previous data then just skip this initial call */
   if (!PrevProcCount) return (0);

   /**************************/
   /* loop through processes */
   /**************************/

   for (cnt = 0; cnt < PROC_MAX; cnt++)
       TopCpuTim[cnt] = TopBioCnt[cnt] = TopDioCnt[cnt] =
          TopPageFlts[cnt] = TopSekTim[cnt] = NULL;

   for (spicnt = 0; spicnt < ProcCount; spicnt++, spiptr++)
   {
      if (Debug)
         fprintf (stdout,
"SPI %2d. %8.8X |%*.*s| cpu:%u pgflt:%u dio:%u bio:%u\n",
                  spicnt+1, spiptr->getspi$l_proc_epid,
                  *spiptr->getspi$o_proc_lname,
                  *spiptr->getspi$o_proc_lname,
                  spiptr->getspi$o_proc_lname+1,
                  spiptr->getspi$l_proc_cputim,
                  spiptr->getspi$l_proc_pageflts,
                  spiptr->getspi$l_proc_diocnt,
                  spiptr->getspi$l_proc_biocnt);

      /*************************/
      /* find in previous data */
      /*************************/

      sprptr = prvptr;
      for (sprcnt = 0; sprcnt < PrevProcCount; sprcnt++, sprptr++)
         if (spiptr->getspi$l_proc_epid == sprptr->getspi$l_proc_epid) break;
      /* if process not found in previous data */
      if (sprcnt >= PrevProcCount) continue;

      if (Debug)
         fprintf (stdout,
"PRV %2d. %8.8X |%*.*s| cpu:%u pgflt:%u dio:%u bio:%u\n",
                  sprcnt+1, sprptr->getspi$l_proc_epid,
                  *spiptr->getspi$o_proc_lname,
                  *spiptr->getspi$o_proc_lname,
                  sprptr->getspi$o_proc_lname+1,
                  sprptr->getspi$l_proc_cputim,
                  sprptr->getspi$l_proc_pageflts,
                  sprptr->getspi$l_proc_diocnt,
                  sprptr->getspi$l_proc_biocnt);

      /********************/
      /* calculate deltas */
      /********************/

      sprptr->getspi$l_proc_cputim =
        (int)((float)(spiptr->getspi$l_proc_cputim -
                      sprptr->getspi$l_proc_cputim) / DeltaSeconds);
      sprptr->getspi$l_proc_biocnt =
         (int)((float)(spiptr->getspi$l_proc_biocnt -
                       sprptr->getspi$l_proc_biocnt) / DeltaSeconds);
      sprptr->getspi$l_proc_diocnt =
         (int)((float)(spiptr->getspi$l_proc_diocnt -
                       sprptr->getspi$l_proc_diocnt) / DeltaSeconds);
      sprptr->getspi$l_proc_pageflts =
         (int)((float)(spiptr->getspi$l_proc_pageflts -
                       sprptr->getspi$l_proc_pageflts) / DeltaSeconds);
      sprptr->getspi$l_proc_SEK = spiptr->getspi$l_proc_kernel_counter -
                                    sprptr->getspi$l_proc_kernel_counter;
      sprptr->getspi$l_proc_SEK += spiptr->getspi$l_proc_executive_counter -
                                     sprptr->getspi$l_proc_executive_counter;
      sprptr->getspi$l_proc_SEK += spiptr->getspi$l_proc_supervisor_counter -
                                     sprptr->getspi$l_proc_supervisor_counter;
      sprptr->getspi$l_proc_SEK =
         (int)((float)sprptr->getspi$l_proc_SEK / DeltaSeconds);

      /**************************/
      /* populate sorted values */
      /**************************/

      if (sprptr->getspi$l_proc_cputim)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopCpuTim[topcnt])
            {
               if (sprptr->getspi$l_proc_cputim >
                   TopCpuTim[topcnt]->getspi$l_proc_cputim)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopCpuTim[cnt] = TopCpuTim[cnt-1];
                  TopCpuTim[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopCpuTim[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_biocnt)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopBioCnt[topcnt])
            {
               if (sprptr->getspi$l_proc_biocnt >
                   TopBioCnt[topcnt]->getspi$l_proc_biocnt)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopBioCnt[cnt] = TopBioCnt[cnt-1];
                  TopBioCnt[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopBioCnt[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_diocnt)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopDioCnt[topcnt])
            {
               if (sprptr->getspi$l_proc_diocnt >
                   TopDioCnt[topcnt]->getspi$l_proc_diocnt)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopDioCnt[cnt] = TopDioCnt[cnt-1];
                  TopDioCnt[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopDioCnt[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_pageflts)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopPageFlts[topcnt])
            {
               if (sprptr->getspi$l_proc_pageflts >
                   TopPageFlts[topcnt]->getspi$l_proc_pageflts)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopPageFlts[cnt] = TopPageFlts[cnt-1];
                  TopPageFlts[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopPageFlts[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_SEK)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopSekTim[topcnt])
            {
               if (sprptr->getspi$l_proc_SEK >
                   TopSekTim[topcnt]->getspi$l_proc_SEK)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopSekTim[cnt] = TopSekTim[cnt-1];
                  TopSekTim[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopSekTim[topcnt] = sprptr;
               break;
            }
         }

      /********************/
      /* end process loop */
      /********************/
   }

   /********************/
   /* write html table */
   /********************/

   zptr = (sptr = sysptr->ProcessBuffer) + sizeof(sysptr->ProcessBuffer)-64;

   sptr += sprintf (sptr, "the.mdsMonProc(\"");
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   for (cnt = RowCount = 0; cnt < PROC_MAX; cnt++)
   {
      if (!(TopCpuTim[cnt] || TopSekTim[cnt] ||
            TopPageFlts[cnt] || TopBioCnt[cnt] || TopDioCnt[cnt]))
         continue;
      RowCount++;

      sptr += sprintf (sptr, "<tr>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopCpuTim[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %1.2f</td>",
                          TopCpuTim[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopCpuTim[cnt]->getspi$o_proc_lname),
                          (float)TopCpuTim[cnt]->getspi$l_proc_cputim/100.0);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopSekTim[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %1.2f</td>",
                          TopSekTim[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopSekTim[cnt]->getspi$o_proc_lname),
                          (float)TopSekTim[cnt]->getspi$l_proc_SEK/100.0);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopPageFlts[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopPageFlts[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopPageFlts[cnt]->getspi$o_proc_lname),
                          TopPageFlts[cnt]->getspi$l_proc_pageflts);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopBioCnt[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopBioCnt[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopBioCnt[cnt]->getspi$o_proc_lname),
                          TopBioCnt[cnt]->getspi$l_proc_biocnt);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopDioCnt[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopDioCnt[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopDioCnt[cnt]->getspi$o_proc_lname),
                          TopDioCnt[cnt]->getspi$l_proc_diocnt);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sptr += sprintf (sptr, "</tr>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
   }

   sptr += sprintf (sptr, "\");");
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   if (RowCount)
      sysptr->ProcessBufferLength = sptr - sysptr->ProcessBuffer;
   else
      sysptr->ProcessBuffer[sysptr->ProcessBufferLength = 0] = '\0';

   if (Debug)
      fprintf (stdout, "\n%d\n%s\n",
               sysptr->ProcessBufferLength, sysptr->ProcessBuffer);

   return (ProcCount);
}

/*****************************************************************************/
/*
Collect process System Performance Information for older VMSes.
Functionality is as per CollectProcSPI1().
*/ 

int CollectProcSPI2 (SYSTEM_DATA *sysptr)

{
#define GETSPI$_PROC 4123

/* structure according to the book! */

typedef struct _getspi$proc_class {
    int getspi$l_proc_ipid;
    int getspi$l_proc_uic;
    short int getspi$w_proc_state;
    char getspi$b_proc_pri;
    char getspi$o_proc_lname [16];
    int getspi$l_proc_gpgcnt;
    int getspi$l_proc_ppgcnt;
    int getspi$l_proc_sts;
    int getspi$l_proc_diocnt;
    int getspi$l_proc_pageflts;
    int getspi$l_proc_cputim;
    int getspi$l_proc_biocnt;
    int getspi$l_proc_epid;
    int getspi$l_proc_efwm;
    int getspi$l_proc_rbstran;
    } GETSPI$PROC_CLASS;

#define PROC_DATA_SIZE sizeof(unsigned long) + \
                       (sizeof(GETSPI$PROC_CLASS)*PROC_CLASS_MAX)

   static int  BufferNumber,
               DiskCount;
   static unsigned short  PrevMinute = -1;
   static char  ProcData1 [PROC_DATA_SIZE],
                ProcData2 [PROC_DATA_SIZE];

   /* initialize structure for SPI items */
   static struct
   {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList1[] =
   {
       { sizeof(ProcData1), GETSPI$_PROC, 0, 0 },
       {0,0,0,0}
   };

   int  cnt, spicnt, sprcnt, status, topcnt,
        PrevProcCount,
        ProcCount,
        RowCount;
   float  tfloat;
   char  *sptr, *zptr;
   unsigned char  ch;
   struct STRUCT_IOSBLK  IOsb;
   GETSPI$PROC_CLASS  *prvptr, *spiptr, *sprptr;
   GETSPI$PROC_CLASS  *TopCpuTim [PROC_MAX],
                      *TopBioCnt [PROC_MAX],
                      *TopDioCnt [PROC_MAX],
                      *TopPageFlts [PROC_MAX];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectProcSPI2()\n");

   sysptr->ProcessBuffer[sysptr->ProcessBufferLength = 0] = '\0';

   /* if process not being monitored */
   if (!(sysptr->ProcCount && sysptr->ProcMonitor)) return (0);

   /***************/
   /* collect SPI */
   /***************/

   /* swap back and forth between the two buffers */
   if (BufferNumber++ % 2)
   {
      spiptr = (GETSPI$PROC_CLASS*)ProcData2;
      prvptr = (GETSPI$PROC_CLASS*)ProcData1;
   }
   else
   {
      spiptr = (GETSPI$PROC_CLASS*)ProcData1;
      prvptr = (GETSPI$PROC_CLASS*)ProcData2;
   }
   ItemList1[0].BufferPtr = spiptr;

   /* collect System Performance Information */
   status = exe$getspi (SpiEfn, 0, 0, &ItemList1, &IOsb, 0, 0);

   if (Debug)
      fprintf (stdout, "exe$getspi() %%X%08.08X  IOsb: %%X%08.08X proc: %d\n",
               status, IOsb.iosb$w_status, *(unsigned long*)spiptr);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) return (0);

   if (sysptr->ProcSanityCheck == 1)
   {
      /* sanity check the data structure */
      sysptr->ProcSanityCheck = 2;
      ProcCount = *(unsigned long*)ItemList1[0].BufferPtr;
      spiptr = (GETSPI$PROC_CLASS*)((char*)spiptr + sizeof(unsigned long));
      for (spicnt = 0; spicnt < ProcCount; spicnt++, spiptr++)
      {
         /* if any seem out-of-range the offsets are probably wrong - hmmm! */
         if (spiptr->getspi$l_proc_epid == 0 ||
             spiptr->getspi$w_proc_state < 0 ||
             spiptr->getspi$w_proc_state > 255 ||
             *spiptr->getspi$o_proc_lname < 0 ||
             *spiptr->getspi$o_proc_lname > 15)
            EXIT_FI_LI (SS$_BUGCHECK);
      }
   }

   /***********************/
   /* get proceess counts */
   /***********************/

   /* number of processes read by SPI */
   ProcCount = *(unsigned long*)ItemList1[0].BufferPtr;
   if (ProcCount == PROC_CLASS_MAX) EXIT_FI_LI (SS$_BUGCHECK);
   /* now step over the process count */
   spiptr = (GETSPI$PROC_CLASS*)((char*)spiptr + sizeof(unsigned long));

   /* number of process read into the previous SPI */
   PrevProcCount = *(unsigned long*)prvptr;
   /* now step over the process count */
   prvptr = (GETSPI$PROC_CLASS*)((char*)prvptr + sizeof(unsigned long));
   /* if there is no previous data then just skip this initial call */
   if (!PrevProcCount) return (0);

   /**************************/
   /* loop through processes */
   /**************************/

   for (cnt = 0; cnt < PROC_MAX; cnt++)
       TopCpuTim[cnt] = TopBioCnt[cnt] = TopDioCnt[cnt] =
          TopPageFlts[cnt] = NULL;

   for (spicnt = 0; spicnt < ProcCount; spicnt++, spiptr++)
   {
      if (Debug)
         fprintf (stdout,
"SPI %2d. %8.8X |%*.*s| cpu:%u pgflt:%u dio:%u bio:%u\n",
                  spicnt+1, spiptr->getspi$l_proc_epid,
                  *spiptr->getspi$o_proc_lname,
                  *spiptr->getspi$o_proc_lname,
                  spiptr->getspi$o_proc_lname+1,
                  spiptr->getspi$l_proc_cputim,
                  spiptr->getspi$l_proc_pageflts,
                  spiptr->getspi$l_proc_diocnt,
                  spiptr->getspi$l_proc_biocnt);

      /*************************/
      /* find in previous data */
      /*************************/

      sprptr = prvptr;
      for (sprcnt = 0; sprcnt < PrevProcCount; sprcnt++, sprptr++)
         if (spiptr->getspi$l_proc_epid == sprptr->getspi$l_proc_epid) break;
      /* if process not found in previous data */
      if (sprcnt >= PrevProcCount) continue;

      if (Debug)
         fprintf (stdout,
"PRV %2d. %8.8X |%*.*s| cpu:%u pgflt:%u dio:%u bio:%u\n",
                  sprcnt+1, sprptr->getspi$l_proc_epid,
                  *spiptr->getspi$o_proc_lname,
                  *spiptr->getspi$o_proc_lname,
                  sprptr->getspi$o_proc_lname+1,
                  sprptr->getspi$l_proc_cputim,
                  sprptr->getspi$l_proc_pageflts,
                  sprptr->getspi$l_proc_diocnt,
                  sprptr->getspi$l_proc_biocnt);

      /********************/
      /* calculate deltas */
      /********************/

      sprptr->getspi$l_proc_cputim =
        (int)((float)(spiptr->getspi$l_proc_cputim -
                      sprptr->getspi$l_proc_cputim) / DeltaSeconds);
      sprptr->getspi$l_proc_biocnt =
         (int)((float)(spiptr->getspi$l_proc_biocnt -
                       sprptr->getspi$l_proc_biocnt) / DeltaSeconds);
      sprptr->getspi$l_proc_diocnt =
         (int)((float)(spiptr->getspi$l_proc_diocnt -
                       sprptr->getspi$l_proc_diocnt) / DeltaSeconds);
      sprptr->getspi$l_proc_pageflts =
         (int)((float)(spiptr->getspi$l_proc_pageflts -
                       sprptr->getspi$l_proc_pageflts) / DeltaSeconds);

      /**************************/
      /* populate sorted values */
      /**************************/

      if (sprptr->getspi$l_proc_cputim)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopCpuTim[topcnt])
            {
               if (sprptr->getspi$l_proc_cputim >
                   TopCpuTim[topcnt]->getspi$l_proc_cputim)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopCpuTim[cnt] = TopCpuTim[cnt-1];
                  TopCpuTim[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopCpuTim[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_biocnt)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopBioCnt[topcnt])
            {
               if (sprptr->getspi$l_proc_biocnt >
                   TopBioCnt[topcnt]->getspi$l_proc_biocnt)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopBioCnt[cnt] = TopBioCnt[cnt-1];
                  TopBioCnt[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopBioCnt[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_diocnt)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopDioCnt[topcnt])
            {
               if (sprptr->getspi$l_proc_diocnt >
                   TopDioCnt[topcnt]->getspi$l_proc_diocnt)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopDioCnt[cnt] = TopDioCnt[cnt-1];
                  TopDioCnt[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopDioCnt[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_pageflts)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopPageFlts[topcnt])
            {
               if (sprptr->getspi$l_proc_pageflts >
                   TopPageFlts[topcnt]->getspi$l_proc_pageflts)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopPageFlts[cnt] = TopPageFlts[cnt-1];
                  TopPageFlts[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopPageFlts[topcnt] = sprptr;
               break;
            }
         }

      /********************/
      /* end process loop */
      /********************/
   }

   /********************/
   /* write html table */
   /********************/

   zptr = (sptr = sysptr->ProcessBuffer) + sizeof(sysptr->ProcessBuffer)-64;

   sptr += sprintf (sptr, "the.mdsMonProc(\"");
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   for (cnt = RowCount = 0; cnt < PROC_MAX; cnt++)
   {
      if (!(TopCpuTim[cnt] || TopPageFlts[cnt] ||
            TopBioCnt[cnt] || TopDioCnt[cnt]))
         continue;
      RowCount++;

      sptr += sprintf (sptr, "<tr>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopCpuTim[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %1.2f</td><td>(n/a)</td>",
                          TopCpuTim[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopCpuTim[cnt]->getspi$o_proc_lname),
                          (float)TopCpuTim[cnt]->getspi$l_proc_cputim/100.0);
      else
         sptr += sprintf (sptr, "<td></td><td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopPageFlts[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopPageFlts[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopPageFlts[cnt]->getspi$o_proc_lname),
                          TopPageFlts[cnt]->getspi$l_proc_pageflts);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopBioCnt[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopBioCnt[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopBioCnt[cnt]->getspi$o_proc_lname),
                          TopBioCnt[cnt]->getspi$l_proc_biocnt);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopDioCnt[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopDioCnt[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopDioCnt[cnt]->getspi$o_proc_lname),
                          TopDioCnt[cnt]->getspi$l_proc_diocnt);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sptr += sprintf (sptr, "</tr>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
   }

   sptr += sprintf (sptr, "\");");
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   if (RowCount)
      sysptr->ProcessBufferLength = sptr - sysptr->ProcessBuffer;
   else
      sysptr->ProcessBuffer[sysptr->ProcessBufferLength = 0] = '\0';

   if (Debug)
      fprintf (stdout, "\n%d\n%s\n",
               sysptr->ProcessBufferLength, sysptr->ProcessBuffer);

   return (ProcCount);
}

/*****************************************************************************/
/*
HTML-escape any forbidden characters in the process name.
Process name is a maximum fifteen character counted string.
Static buffer so can only be called the once per print call.
*/

char* MungeProcName (char *ProcName)

{
   /* maximum fifteen ampersand's worth */
   static char  NameBuffer [80];

   int  cnt, filcnt;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   sptr = NameBuffer;
   cnt = *ProcName;
   filcnt = 15 - cnt;
   cptr = ProcName + 1;
   while (cnt--)
   {
      switch (*cptr)
      {
         case '<' :
            *(ULONGPTR)sptr = '&lt;';
            sptr += 4;
            break;
         case '>' :
            *(ULONGPTR)sptr = '&gt;';
            sptr += 4;
            break;
         case '&' :
            *(ULONGPTR)sptr = '&amp';
            sptr += 4;
            *sptr++ = ';';
            break;
         default :
            *sptr++ = *cptr;
      }
      cptr++;
   }
   while (filcnt--) *sptr++ = ' ';
   *sptr = '\0';

   return (NameBuffer);
}

/*****************************************************************************/
/*
Sets up a WebSocket client which then runs independently and asynchronously in
a CGIplus environment.
*/

void WebSockRequest ()

{
   char  *cptr;
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   /* allocate a data structure for this connected client */
   clptr = calloc (1, sizeof(CLIENT_DATA));
   if (!clptr) EXIT_FI_LI (vaxc$errno);

   /* allocate a WebSocket library structure */
   if (!(clptr->WsLibPtr = WsLibCreate (clptr, WebSockDestroy)))
   {
      /* failed, commonly on some WebSocket protocol issue */
      free (clptr);
      return;
   }

   WebSockConnectedCount++;

   /* open the WebSocket IPC to the server (and from there to the client) */
   WsLibOpen (clptr->WsLibPtr);

   WsLibRead (clptr->WsLibPtr,
              clptr->InputBuffer,
              sizeof(clptr->InputBuffer)-1,
              WebSockReadAst);
}

/*****************************************************************************/
/*
WebSocket has closed, request concluded.
*/

void WebSockDestroy (struct WsLibStruct *wsptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WebSockDestroy()\n");

   free (WsLibGetUserData (wsptr));

   if (WebSockConnectedCount) WebSockConnectedCount--;
}

/*****************************************************************************/
/*
For MONDESI this is really just a demonstration of the WebSocket asynchronous
client to server communication, where the cluster statistics are turned on or
off (saving a small amount of bandwidth) depending on whether they are visible
or not.
*/

void WebSockReadAst (struct WsLibStruct *wsptr)

{
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "WebSockReadAst() %%X%08.08X\n",
               WsLibReadStatus (wsptr));

   if (VMSnok (WsLibReadStatus (wsptr)))
   {
      WsLibClose (wsptr, 0, NULL);
      return;
   }

   clptr = WsLibGetUserData (wsptr);

   /* ensure it's null-terminated */
   clptr->InputBuffer[WsLibReadCount(wsptr)] = '\0';

   if (!strncmp (clptr->InputBuffer, "procmon=", 8))
      clptr->ProcMonitor = atoi(clptr->InputBuffer+8);
   else
      sprintf (NotePad, "$[%d] |%16.16s| %04.04x%04.04x%04.04x",
               __LINE__, clptr->InputBuffer,
               ntohl(*((unsigned long*)clptr->InputBuffer)),
               ntohl(*((unsigned long*)(clptr->InputBuffer+4))),
               ntohl(*((unsigned long*)(clptr->InputBuffer+8))));

   /* read next */
   WsLibRead (wsptr,
              clptr->InputBuffer,
              sizeof(clptr->InputBuffer)-1,
              WebSockReadAst);
}

/*****************************************************************************/
/*
Decode URL-encoded string.  Resultant string is always the same size or smaller
so it can be done in-situ!  Returns the size of the resultant string.
*/

int UrlDecode (char *String)

{
   unsigned char  ch;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UrlDecode() |%s|\n", String);

   cptr = sptr = String;
   while (*cptr)
   {
      switch (*cptr)
      {
         case '+' :

            *sptr++ = ' ';
            cptr++;

            continue;

         case '%' :

            cptr++;
            while (*cptr == '\r' || *cptr == '\n') cptr++;
            ch = 0;
            if (*cptr >= '0' && *cptr <= '9')
               ch = (*cptr - (int)'0') << 4;
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               ch = (toupper(*cptr) - (int)'A' + 10) << 4;
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            while (*cptr == '\r' || *cptr == '\n') cptr++;
            if (*cptr >= '0' && *cptr <= '9')
               ch += (*cptr - (int)'0');
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               ch += (toupper(*cptr) - (int)'A' + 10);
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            *sptr++ = ch;

            continue;

         case '\r' :
         case '\n' :

            cptr++;
            continue;

         default :
            
            *sptr++ = *cptr++;

      }
   }
   *sptr = '\0';

   if (Debug) fprintf (stdout, "|%s|\n", String);
   return (sptr-String);
}

/****************************************************************************/

